from sqlalchemy.testing.requirements import SuiteRequirements

from sqlalchemy.testing import exclusions


class Requirements(SuiteRequirements):

    @property
    def reflects_pk_names(self):
        return exclusions.open()

    @property
    def index_reflection(self):
        return exclusions.closed()

    @property
    def unique_constraint_reflection(self):
        """target dialect supports reflection of unique constraints"""
        return exclusions.closed()

    @property
    def self_referential_foreign_keys(self):
        """Target database must support self-referential foreign keys."""
        return exclusions.open()

    @property
    def implicitly_named_constraints(self):
        """target database must apply names to unnamed constraints."""
        return exclusions.open()

    @property
    def returning(self):
        """target platform supports RETURNING."""
        return exclusions.closed()

    @property
    def empty_strings_varchar(self):
        """target database can persist/return an empty string with a
        varchar. """
        return exclusions.closed()

    @property
    def text_type(self):
        """Target database must support an unbounded Text()
        type such as TEXT or CLOB """
        return exclusions.closed()

    @property
    def datetime_microseconds(self):
        """target dialect supports representation of Python
        datetime.datetime() with microsecond objects."""
        return exclusions.closed()

    @property
    def datetime_historic(self):
        """target dialect supports representation of Python
        datetime.datetime() objects with historic (pre 1970) values."""
        return exclusions.open()

    @property
    def date_historic(self):
        """target dialect supports representation of Python
        datetime.date() objects with historic (pre 1970) values."""
        return exclusions.open()

    @property
    def time(self):
        """target dialect supports representation of Python
        datetime.time() objects."""
        return exclusions.closed()

    @property
    def time_microseconds(self):
        """target dialect supports representation of Python
        datetime.time() with microsecond objects."""
        return exclusions.closed()

    @property
    def unbounded_varchar(self):
        """Target database must support VARCHAR with no length"""
        return exclusions.closed()

    @property
    def implements_get_lastrowid(self):
        """"target dialect implements the executioncontext.get_lastrowid()
            method without reliance on RETURNING."""
        return exclusions.closed()

    @property
    def emulated_lastrowid(self):
        """"target dialect retrieves cursor.lastrowid, or fetches
        from a database-side function after an insert() construct executes,
        within the get_lastrowid() method.

        Only dialects that "pre-execute", or need RETURNING to get last
        inserted id, would return closed/fail/skip for this.
        """
        return exclusions.open()

    @property
    def schemas(self):
        """Target database supports named schemas
        """
        return exclusions.open()

    @property
    def views(self):
        """Target database must support VIEWs."""

        return exclusions.open()

    @property
    def view_reflection(self):
        """Target database supports view metadata
        """
        return exclusions.open()

    @property
    def precision_numerics_enotation_large(self):
        """Dialect converts small/large scale decimals into scientific notation
        """
        return exclusions.open()

    @property
    def temporary_tables(self):
        """target database supports temporary tables"""
        return exclusions.closed()

    @property
    def temp_table_names(self):
        """target dialect supports listing of temporary table names"""
        return exclusions.closed()

    @property
    def temp_table_reflection(self):
        return exclusions.closed()

    @property
    def offset(self):
        """target database can render OFFSET, or an equivalent, in a
        SELECT.
        """
        return exclusions.closed()

    @property
    def order_by_col_from_union(self):
        """target database supports ordering by a column from a SELECT
           inside of a UNION
           E.g.  (SELECT id, ...) UNION (SELECT id, ...) ORDER BY id """
        return exclusions.open()

    @property
    def bound_limit_offset(self):
        """target database can render LIMIT and/or OFFSET using a bound
        parameter
        """
        return exclusions.closed()

    @property
    def duplicate_key_raises_integrity_error(self):
        return exclusions.closed()

    @property
    def independent_connections(self):
        return exclusions.open()

    @property
    def parens_in_union_contained_select_w_limit_offset(self):
        return exclusions.closed()

    @property
    def parens_in_union_contained_select_wo_limit_offset(self):
        return exclusions.closed()

    @property
    def broken_cx_oracle6_numerics(config):
        return exclusions.closed()

    @property
    def cross_schema_fk_reflection(self):
        return exclusions.closed()

    @property
    def ctes(self):
        """Target database supports CTEs"""
        """Can't be opened as CTE tests require DB support for 'WITH RECURSIVE'
           not supported by EXASOL"""
        return exclusions.closed()

    @property
    def standalone_null_binds_whereclause(self):
        """target database/driver supports bound parameters with NULL in the
        WHERE clause, in situations where it has to be typed.
        """
        return exclusions.closed()



try:
    from ._version import version as __version__
except ImportError:  # pragma: no cover
    __version__ = "unknown"

from sqlalchemy_exasol import base, pyodbc

# default dialect
base.dialect = pyodbc.dialect

__all__ = (
    'BLOB', 'BOOLEAN', 'CHAR', 'DATE', 'DATETIME', 'DECIMAL', 'FLOAT', 'INTEGER',
    'NUMERIC', 'SMALLINT', 'TEXT', 'TIME', 'TIMESTAMP', 'VARCHAR', 'dialect', 'REAL'
)



import datetime
import six

from sqlalchemy_exasol.pyodbc import EXADialect_pyodbc
from sqlalchemy_exasol import base


def raw_sql(query):
    """
    Converts an SQLAlchemy object to a single raw SQL string. Note
    that this function currently only supports Exasol queries!

    :param query: An SQLAlchemy query object

    :returns: A string of raw SQL
    :rtype: string
    """
    dialect = EXADialect_pyodbc()

    class LiteralCompiler(base.EXACompiler):
        def visit_bindparam(self, bindparam,
                            within_columns_clause=False,
                            literal_binds=False, **kwargs):
            return super(LiteralCompiler, self).render_literal_bindparam(
                bindparam, within_columns_clause=within_columns_clause,
                literal_binds=literal_binds, **kwargs)

        def render_literal_value(self, value, type_):
            if value is None:
                return 'NULL'
            elif isinstance(value, six.binary_type):
                return u"'{value}'".format(value=value.decode('utf-8'))
            elif isinstance(value, six.text_type):
                return u"'{value}'".format(value=value)
            elif type(value) is datetime.date:
                return "to_date('{value}', 'YYYY-MM-DD')"\
                    .format(value=value.strftime('%Y-%m-%d'))
            elif type(value) is datetime.datetime:
                return u"to_timestamp('{value}', 'YYYY-MM-DD HH24:MI:SS.FF6')"\
                    .format(value=value.strftime('%Y-%m-%d %H:%M:%S.%f'))
            else:
                return u"{}".format(value)

    compiler = LiteralCompiler(dialect, query)
    return compiler.string



# import all SQLAlchemy tests for this dialect
import pytest
from sqlalchemy.testing.suite import *  # noqa: F403, F401

from sqlalchemy.testing.suite import CompoundSelectTest \
    as _CompoundSelectTest
from sqlalchemy.testing.suite import ExpandingBoundInTest \
    as _ExpandingBoundInTest
from sqlalchemy.testing.suite import NumericTest \
    as _NumericTest
from sqlalchemy.testing.suite import QuotedNameArgumentTest \
    as _QuotedNameArgumentTest


class CompoundSelectTest(_CompoundSelectTest):

    """ Skip this test as EXASOL does not allow EXISTS or IN predicates
        as part of the select list. Skipping is implemented by redefining
        the method as proposed by SQLAlchemy docs for new dialects."""
    def test_null_in_empty_set_is_false(self):
        return


class ExpandingBoundInTest(_ExpandingBoundInTest):

    """ Skip this test as EXASOL does not allow EXISTS or IN predicates
        as part of the select list. Skipping is implemented by redefining
        the method as proposed by SQLAlchemy docs for new dialects."""
    def test_null_in_empty_set_is_false(self):
        return


class NumericTest(_NumericTest):

    """ FIXME: test skipped to allow upgrading to SQLAlchemy 1.3.x due
        to vulnerability in 1.2.x. Need to understand reason for this.
        Hypothesis is that the data type is not correctly coerced between
        EXASOL and pyodbc."""
    def test_decimal_coerce_round_trip(self):
        return


class QuotedNameArgumentTest(_QuotedNameArgumentTest):

    """ This suite was added to SQLAlchemy 1.3.19 on July 2020 to address
        issues in other dialects related to object names that contain quotes 
        and double quotes. Since this feature is not relevant to the 
        Exasol dialect, the entire suite will be skipped. More info on fix:
        https://github.com/sqlalchemy/sqlalchemy/issues/5456 """
    @pytest.mark.skip()
    def test_get_table_options(self, name):
        return

    @pytest.mark.skip()
    def test_get_view_definition(self, name):
        return

    @pytest.mark.skip()
    def test_get_columns(self, name):
        return

    @pytest.mark.skip()
    def test_get_pk_constraint(self, name):
        return

    @pytest.mark.skip()
    def test_get_foreign_keys(self, name):
        return

    @pytest.mark.skip()
    def test_get_indexes(self, name):
        return

    @pytest.mark.skip()
    def test_get_unique_constraints(self, name):
        return

    @pytest.mark.skip()
    def test_get_table_comment(self, name):
        return

    @pytest.mark.skip()
    def test_get_check_constraints(self, name):
        return


# -*- coding: UTF-8 -*-
import datetime

from sqlalchemy import MetaData, Table, Column, Integer, String, Date, DateTime
from sqlalchemy import or_, select, literal_column
from sqlalchemy import testing, inspect
from sqlalchemy.schema import DropConstraint, AddConstraint
from sqlalchemy.testing import fixtures, config

from sqlalchemy_exasol.base import EXAExecutionContext
from sqlalchemy_exasol.constraints import DistributeByConstraint
from sqlalchemy_exasol.merge import merge
from sqlalchemy_exasol.util import raw_sql


class MergeTest(fixtures.TablesTest):

    #run_deletes = 'each'

    __backend__ = True

    __engine_options__ = {"implicit_returning": False}

    @classmethod
    def define_tables(cls, metadata):
        Table('t', metadata,
            Column('id', Integer),
            Column('name', String(20)),
            Column('age', Integer)
        )
        Table('s', metadata,
            Column('id', Integer),
            Column('age', Integer)
        )

    def test_merge_update(self):
        t = self.tables.t
        s = self.tables.s

        config.db.execute(
            t.insert(),
            [dict(id=1, age=None, name='Ulf')]
        )
        config.db.execute(
            s.insert(),
            [dict(id=1, age=10)]
        )

        m = merge(t, s, t.c.id == s.c.id).update()
        mr = config.db.execute(m)

        r = config.db.execute(t.select()).first()
        assert r.age == 10

    def test_merge_update_where(self):
        t = self.tables.t
        s = self.tables.s

        config.db.execute(
            t.insert(),
            [
                dict(id=1, name='Ulf'),
                dict(id=2, name='Karl'),
            ]
        )
        config.db.execute(
            s.insert(),
            [
                dict(id=1, age=10),
                dict(id=2, age=20),
                dict(id=3, age=30),
            ]
        )

        m = merge(t, s, t.c.id == s.c.id).update(
                    values={t.c.age: s.c.age},
                    where=t.c.id == s.c.id
                )
        mr = config.db.execute(m)

        r = config.db.execute(t.select().where(t.c.id == 1)).first()
        assert r.age == 10
        r = config.db.execute(t.select().where(t.c.id == 2)).first()
        assert r.age == 20
        r = config.db.execute(t.select().where(t.c.id == 3)).fetchall()
        assert len(r) == 0

    def test_merge_insert(self):
        t = self.tables.t
        s = self.tables.s

        config.db.execute(
            t.insert(),
            [
                dict(id=1, name='Ulf'),
            ]
        )
        config.db.execute(
            s.insert(),
            [
                dict(id=1, age=10),
                dict(id=2, age=20),
            ]
        )

        m = merge(t, s, t.c.id == s.c.id).insert()
        config.db.execute(m)
        r = config.db.execute(t.select()).fetchall()
        assert len(r) == 2

    def test_merge_delete_where(self):
        t = self.tables.t
        s = self.tables.s

        config.db.execute(
            t.insert(),
            [
                dict(id=1, name='Ulf'),
            ]
        )
        config.db.execute(
            s.insert(),
            [
                dict(id=1, age=10),
                dict(id=2, age=20),
                dict(id=3, age=2),
            ]
        )

        m = merge(t, s, t.c.id == s.c.id).insert(
                where=s.c.age > 5
            )
        config.db.execute(m)
        r = config.db.execute(t.select()).fetchall()
        assert len(r) == 2

    def test_merge_delete(self):
        t = self.tables.t
        s = self.tables.s

        config.db.execute(
            t.insert(),
            [
                dict(id=1, name='Ulf'),
                dict(id=4, name='Peter'),
            ]
        )
        config.db.execute(
            s.insert(),
            [
                dict(id=1, age=10),
                dict(id=2, age=20),
            ]
        )

        m = merge(t, s, t.c.id == s.c.id).delete()
        config.db.execute(m)
        r = config.db.execute(t.select()).fetchall()
        assert len(r) == 1

    def test_merge_insert_computed_column(self):
        t = self.tables.t
        s = self.tables.s

        config.db.execute(t.insert(), [dict(id=1, name="Ulf")])
        config.db.execute(s.insert(), [dict(id=1, age=10), dict(id=2, age=20)])

        source_expr = select(
            [
                (s.c.id * literal_column("20")).label("id"),
                (s.c.age * literal_column("20")).label("age"),
                literal_column("'Bernd'").label("name"),
            ]
        ).alias("source")

        # Expecting both rows of source_expr to be inserted
        m = merge(t, source_expr, t.c.id == source_expr.c.id).insert()
        config.db.execute(m)
        results = config.db.execute(t.select().where(t.c.name == "Bernd")).fetchall()
        assert len(results) == 2

        for r in results:
            assert r.id > 10
            assert r.age >= 10
            assert r.name == "Bernd"

    def test_merge_update_computed_column(self):
        t = self.tables.t
        s = self.tables.s

        config.db.execute(t.insert(), [dict(id=1, name="Ulf")])
        config.db.execute(s.insert(), [dict(id=1, age=10), dict(id=2, age=20)])

        # Only id=1 will match, so only one row should be updated.
        source_expr = select(
            [
                s.c.id.label("id"),
                (s.c.age * literal_column("20")).label("age"),
                literal_column("'Bernd'").label("name"),
            ]
        ).alias("source")

        m = merge(t, source_expr, t.c.id == source_expr.c.id).update()
        mr = config.db.execute(m)

        results = config.db.execute(t.select().where(t.c.name == "Bernd")).fetchall()
        assert len(results) == 1
        r = results[0]
        assert r.age == 200
        assert r.name == "Bernd"

    def test_merge_update_insert(self):
        t = self.tables.t
        s = self.tables.s

        config.db.execute(
            t.insert(),
            [
                dict(id=1, name='Ulf'),
            ]
        )
        config.db.execute(
            s.insert(),
            [
                dict(id=1, age=10),
                dict(id=2, age=20),
            ]
        )

        m = merge(t, s, t.c.id == s.c.id).update(
                values={t.c.age: s.c.age}
            ).insert()
        config.db.execute(m)
        r = config.db.execute(t.select()).fetchall()
        assert len(r) == 2

    def test_merge_update_delete(self):
        t = self.tables.t
        s = self.tables.s

        config.db.execute(
            t.insert(),
            [dict(id=1, name='Ulf')]
        )
        config.db.execute(
            s.insert(),
            [dict(id=1, age=10)]
        )

        m = merge(t, s, t.c.id == s.c.id).update(
                values={t.c.age: s.c.age}
            ).delete()
        config.db.execute(m)
        r = config.db.execute(t.select()).fetchall()
        assert len(r) == 0

    def test_merge_update_delete_where(self):
        t = self.tables.t
        s = self.tables.s

        config.db.execute(
            t.insert(),
            [dict(id=1, name='Ulf')]
        )
        config.db.execute(
            s.insert(),
            [dict(id=1, age=10)]
        )

        m = merge(t, s, t.c.id == s.c.id).update(
                values={t.c.age: s.c.age}
            ).delete(t.c.id == s.c.id)
        config.db.execute(m)
        r = config.db.execute(t.select()).fetchall()
        assert len(r) == 0


class DefaultsTest(fixtures.TablesTest):
    __backend__ = True

    @classmethod
    def define_tables(cls, metadata):
        default_date = datetime.date(1900, 1, 1)
        Table('t', metadata,
            Column('id', Integer),
            Column('name', String(20)),
            Column('active_from', Date, default=default_date)
        )

    def test_insert_with_default_value(self):
        t = self.tables.t

        config.db.execute(
            t.insert(),
            [{'name': 'Henrik'}]
        )
        (_, _, active_from) = config.db.execute(t.select()).fetchone()
        assert active_from == datetime.date(1900, 1, 1)

class KeywordTest(fixtures.TablesTest):

    __backend__ = True

    def test_keywords(self):
        keywords = config.db.execute('select distinct(lower(keyword)) as keyword ' +
            'from SYS.EXA_SQL_KEYWORDS where reserved = True order by keyword').fetchall()
        db_keywords = set([k[0] for k in keywords])
        #assert db_keywords <= RESERVED_WORDS

class AutocommitTest(fixtures.TablesTest):
    __backend__ = False

    def test_trunctate(self):
        ctx = EXAExecutionContext()
        assert ctx.should_autocommit_text('truncate test;')

class ConstraintsTest(fixtures.TablesTest):
    __backend__ = True

    @classmethod
    def define_tables(cls, metadata):
        Table('t', metadata,
           Column('a', Integer),
           Column('b', Integer),
           Column('c', Integer),
           DistributeByConstraint('a', 'b')
        )


    def test_distribute_by_constraint(self):
        try:
           reflected = Table('t', MetaData(testing.db), autoload=True)
        except:
           assert False
        #TODO: check that reflected table object is identical
        # i.e. contains the constraint
        insp = inspect(testing.db)
        for c in insp.get_columns('t'):
            if not (c['name'] == 'c'):
                assert c['is_distribution_key'] == True
            else:
                assert c['is_distribution_key'] == False

    def test_alter_table_distribute_by(self):
        dbc = DistributeByConstraint('a', 'b')
        self.tables.t.append_constraint(dbc)

        config.db.execute(DropConstraint(dbc))

        insp = inspect(testing.db)
        for c in insp.get_columns('t'):
            assert c['is_distribution_key'] == False

        config.db.execute(AddConstraint(dbc))

        insp = inspect(testing.db)
        for c in insp.get_columns('t'):
            if not (c['name'] == 'c'):
                assert c['is_distribution_key'] == True
            else:
                assert c['is_distribution_key'] == False


class UtilTest(fixtures.TablesTest):
    __backend__ = True

    __engine_options__ = {"implicit_returning": False}

    @classmethod
    def define_tables(cls, metadata):
        Table('t', metadata,
            Column('id', Integer),
            Column('name', String(20)),
            Column('age', Integer),
            Column('day', Date),
            Column('created', DateTime)
        )

    def test_raw_sql(self):

        restriction = or_(self.tables.t.c.id == 1,
                          self.tables.t.c.name == "bob",
                          self.tables.t.c.day == datetime.date(2017,1,1),
                          self.tables.t.c.created == datetime.datetime(2017,1,1,12,0,0))
        sel = self.tables.t.select().where(restriction)
        sql = """SELECT t.id, t.name, t.age, t."day", t.created 
FROM t 
WHERE t.id = 1 OR t.name = \'bob\' OR t."day" = to_date(\'2017-01-01\', \'YYYY-MM-DD\') OR t.created = to_timestamp(\'2017-01-01 12:00:00.000000\', \'YYYY-MM-DD HH24:MI:SS.FF6\')"""
        assert raw_sql(sel) == sql



# -*- coding: UTF-8 -*-

import pytest
from sqlalchemy import create_engine
from sqlalchemy.engine.url import URL
from sqlalchemy.sql.sqltypes import INTEGER, VARCHAR
from sqlalchemy.testing import fixtures, config

from sqlalchemy_exasol.base import EXADialect

TEST_GET_METADATA_FUNCTIONS_SCHEMA = "test_get_metadata_functions_schema"
ENGINE_NONE_DATABASE = "ENGINE_NONE_DATABASE"
ENGINE_SCHEMA_DATABASE = "ENGINE_SCHEMA_DATABASE"
ENGINE_SCHEMA_2_DATABASE = "ENGINE_SCHEMA_2_DATABASE"


class MetadataTest(fixtures.TablesTest):
    __backend__ = True

    @classmethod
    def define_tables(cls, metadata):
        cls.schema = TEST_GET_METADATA_FUNCTIONS_SCHEMA
        cls.schema_2 = "test_get_metadata_functions_schema_2"
        with config.db.begin() as c:
            try:
                c.execute("DROP SCHEMA %s CASCADE" % cls.schema)
            except Exception as e:
                print(e)
                pass
            c.execute("CREATE SCHEMA %s" % cls.schema)
            c.execute(
                "CREATE TABLE %s.t (pid1 int, pid2 int, name VARCHAR(20), age int, PRIMARY KEY (pid1,pid2))" % cls.schema)
            c.execute(
                "CREATE TABLE {schema}.s (id1 int primary key, fid1 int, fid2 int, age int, CONSTRAINT fk_test FOREIGN KEY (fid1,fid2) REFERENCES {schema}.t(pid1,pid2))".format(
                    schema=cls.schema))
            cls.view_defintion = "CREATE VIEW {schema}.v AS select * from {schema}.t".format(schema=cls.schema)
            c.execute(cls.view_defintion)

            try:
                c.execute("DROP SCHEMA %s CASCADE" % cls.schema_2)
            except Exception as e:
                print(e)
                pass
            c.execute("CREATE SCHEMA %s" % cls.schema_2)
            c.execute(
                "CREATE TABLE %s.t_2 (pid1 int, pid2 int, name VARCHAR(20), age int, PRIMARY KEY (pid1,pid2))" % cls.schema_2)
            c.execute("CREATE VIEW {schema}.v_2 AS select * from {schema}.t_2".format(schema=cls.schema_2))
            c.execute("COMMIT")

            cls.engine_none_database = cls.create_engine_with_database_name(c, None)
            cls.engine_schema_database = cls.create_engine_with_database_name(c, cls.schema)
            cls.engine_schema_2_database = cls.create_engine_with_database_name(c, cls.schema_2)
            cls.engine_map = {
                ENGINE_NONE_DATABASE: cls.engine_none_database,
                ENGINE_SCHEMA_DATABASE: cls.engine_schema_database,
                ENGINE_SCHEMA_2_DATABASE: cls.engine_schema_2_database
            }

    @classmethod
    def generate_url_with_database_name(cls, connection, new_database_name):
        database_url = config.db_url
        new_args = database_url.translate_connect_args()
        new_args["database"] = new_database_name
        new_database_url = URL(drivername=database_url.drivername, query=database_url.query, **new_args)
        return new_database_url

    @classmethod
    def create_engine_with_database_name(cls, connection, new_database_name):
        url = cls.generate_url_with_database_name(connection, new_database_name)
        engine = create_engine(url)
        return engine

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_get_schema_names(self, engine_name, use_sql_fallback):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            schema_names = dialect.get_schema_names(connection=c, use_sql_fallback=use_sql_fallback)
            assert self.schema in schema_names and self.schema_2 in schema_names

    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_compare_get_schema_names_for_sql_and_odbc(self, engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            schema_names_fallback = dialect.get_schema_names(connection=c, use_sql_fallback=True)
            schema_names_odbc = dialect.get_schema_names(connection=c)
            assert sorted(schema_names_fallback) == sorted(schema_names_odbc)

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_get_table_names(self, use_sql_fallback, engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            table_names = dialect.get_table_names(connection=c, schema=self.schema, use_sql_fallback=use_sql_fallback)
            assert "t" in table_names and "s" in table_names

    @pytest.mark.parametrize("schema", [TEST_GET_METADATA_FUNCTIONS_SCHEMA, None])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_compare_get_table_names_for_sql_and_odbc(self, schema, engine_name):
        with self.engine_map[engine_name].begin() as c:
            if schema is None:
                c.execute("OPEN SCHEMA %s" % self.schema)
            dialect = EXADialect()
            table_names_fallback = dialect.get_table_names(connection=c, schema=schema, use_sql_fallback=True)
            table_names_odbc = dialect.get_table_names(connection=c, schema=schema)
            assert table_names_fallback == table_names_odbc

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_has_table_table_exists(self, use_sql_fallback, engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            has_table = dialect.has_table(connection=c, schema=self.schema, table_name="t",
                                          use_sql_fallback=use_sql_fallback)
            assert has_table, "Table %s.T was not found, but should exist" % self.schema

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_has_table_table_exists_not(self, use_sql_fallback, engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            has_table = dialect.has_table(connection=c, schema=self.schema, table_name="not_exist",
                                          use_sql_fallback=use_sql_fallback)
            assert not has_table, "Table %s.not_exist was found, but should not exist" % self.schema

    @pytest.mark.parametrize("schema", [TEST_GET_METADATA_FUNCTIONS_SCHEMA, None])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_compare_has_table_for_sql_and_odbc(self, schema, engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            has_table_fallback = dialect.has_table(connection=c, schema=schema, use_sql_fallback=True, table_name="t")
            has_table_odbc = dialect.has_table(connection=c, schema=schema, table_name="t")
            assert has_table_fallback == has_table_odbc, "Expected table %s.t with odbc and fallback" % schema

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_get_view_names(self, use_sql_fallback,engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            view_names = dialect.get_view_names(connection=c, schema=self.schema, use_sql_fallback=use_sql_fallback)
            assert "v" in view_names

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_get_view_names_for_sys(self, use_sql_fallback, engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            view_names = dialect.get_view_names(connection=c, schema="sys", use_sql_fallback=use_sql_fallback)
            assert len(view_names) == 0

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_get_view_definition(self, use_sql_fallback,engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            view_definition = dialect.get_view_definition(connection=c, schema=self.schema, view_name="v",
                                                          use_sql_fallback=use_sql_fallback)
            assert self.view_defintion == view_definition

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_get_view_definition_view_name_none(self, use_sql_fallback,engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            view_definition = dialect.get_view_definition(connection=c, schema=self.schema, view_name=None,
                                                          use_sql_fallback=use_sql_fallback)
            assert view_definition is None

    @pytest.mark.parametrize("schema", [TEST_GET_METADATA_FUNCTIONS_SCHEMA, None])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_compare_get_view_names_for_sql_and_odbc(self, schema,engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            c.execute("OPEN SCHEMA %s" % self.schema)
            view_names_fallback = dialect.get_view_names(connection=c, schema=schema, use_sql_fallback=True)
            view_names_odbc = dialect.get_view_names(connection=c, schema=schema)
            assert view_names_fallback == view_names_odbc

    @pytest.mark.parametrize("schema", [TEST_GET_METADATA_FUNCTIONS_SCHEMA, None])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_compare_get_view_definition_for_sql_and_odbc(self, schema,engine_name):
        with self.engine_map[engine_name].begin() as c:
            if schema is None:
                c.execute("OPEN SCHEMA %s" % self.schema)
            view_name = "v"
            dialect = EXADialect()
            view_definition_fallback = dialect.get_view_definition(
                connection=c, view_name=view_name, schema=schema, use_sql_fallback=True)
            view_definition_odbc = dialect.get_view_definition(
                connection=c, view_name=view_name, schema=schema)
            assert view_definition_fallback == view_definition_odbc

    @pytest.mark.parametrize("schema", [TEST_GET_METADATA_FUNCTIONS_SCHEMA, None])
    @pytest.mark.parametrize("table", ["t", "s", "unknown"])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_compare_get_columns_for_sql_and_odbc(self, schema, table, engine_name):
        with self.engine_map[engine_name].begin() as c:
            if schema is None:
                c.execute("OPEN SCHEMA %s" % self.schema)
            dialect = EXADialect()
            columns_fallback = dialect.get_columns(connection=c, table_name=table, schema=schema, use_sql_fallback=True)
            columns_odbc = dialect.get_columns(connection=c, table_name=table, schema=schema)
            assert str(columns_fallback) == str(columns_odbc)  # object equality doesn't work for sqltypes

    @pytest.mark.parametrize("schema", [TEST_GET_METADATA_FUNCTIONS_SCHEMA, None])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_compare_get_columns_none_table_for_sql_and_odbc(self, schema, engine_name):
        with self.engine_map[engine_name].begin() as c:
            if schema is None:
                c.execute("OPEN SCHEMA %s" % self.schema)
            dialect = EXADialect()
            table = None
            columns_fallback = dialect.get_columns(connection=c, table_name=table, schema=schema,
                                                       use_sql_fallback=True)
            columns_odbc = dialect.get_columns(connection=c, table_name=table, schema=schema)
            assert str(columns_fallback) == str(columns_fallback)

    def make_columns_comparable(self, column_list):  # object equality doesn't work for sqltypes
        return sorted([{k: str(v) for k, v in column.items()} for column in column_list], key=lambda k: k["name"])

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_get_columns(self, use_sql_fallback, engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            columns = dialect.get_columns(connection=c, schema=self.schema, table_name="t",
                                          use_sql_fallback=use_sql_fallback)
            expected = [{'default': None,
                         'is_distribution_key': False,
                         'name': 'pid1',
                         'nullable': False,
                         'type': INTEGER()},
                        {'default': None,
                         'is_distribution_key': False,
                         'name': 'pid2',
                         'nullable': False,
                         'type': INTEGER()},
                        {'default': None,
                         'is_distribution_key': False,
                         'name': 'name',
                         'nullable': True,
                         'type': VARCHAR(length=20)},
                        {'default': None,
                         'is_distribution_key': False,
                         'name': 'age',
                         'nullable': True,
                         'type': INTEGER()},
                        ]

            assert self.make_columns_comparable(expected) == self.make_columns_comparable(columns)

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_get_columns_table_name_none(self, use_sql_fallback, engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            columns = dialect.get_columns(connection=c, schema=self.schema, table_name=None,
                                              use_sql_fallback=use_sql_fallback)
            assert columns == []

    @pytest.mark.parametrize("schema", [TEST_GET_METADATA_FUNCTIONS_SCHEMA, None])
    @pytest.mark.parametrize("table", ["t", "s"])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_compare_get_pk_constraint_for_sql_and_odbc(self, schema, table, engine_name):
        with self.engine_map[engine_name].begin() as c:
            if schema is None:
                c.execute("OPEN SCHEMA %s" % self.schema)
            dialect = EXADialect()
            pk_constraint_fallback = dialect.get_pk_constraint(connection=c, table_name=table, schema=schema,
                                                               use_sql_fallback=True)
            pk_constraint_odbc = dialect.get_pk_constraint(connection=c, table_name=table, schema=schema)
            assert str(pk_constraint_fallback) == str(pk_constraint_odbc)

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_get_pk_constraint(self, use_sql_fallback, engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            pk_constraint = dialect.get_pk_constraint(connection=c, schema=self.schema, table_name="t",
                                                      use_sql_fallback=use_sql_fallback)
            assert pk_constraint["constrained_columns"] == ['pid1', 'pid2'] and \
                   pk_constraint["name"].startswith("sys_")
    
    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_get_pk_constraint_table_name_none(self, use_sql_fallback, engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            pk_constraint = dialect.get_pk_constraint(connection=c, schema=self.schema, table_name=None,
                                                      use_sql_fallback=use_sql_fallback)
            assert pk_constraint is None

    @pytest.mark.parametrize("table", ["t", "s"])
    @pytest.mark.parametrize("schema", [TEST_GET_METADATA_FUNCTIONS_SCHEMA, None])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_compare_get_foreign_keys_for_sql_and_odbc(self, schema, table, engine_name):
        with self.engine_map[engine_name].begin() as c:
            if schema is None:
                c.execute("OPEN SCHEMA %s" % self.schema_2)
            dialect = EXADialect()
            foreign_keys_fallback = dialect.get_foreign_keys(connection=c, table_name=table, schema=schema,
                                                             use_sql_fallback=True)
            foreign_keys_odbc = dialect.get_foreign_keys(connection=c, table_name=table, schema=schema)
            assert str(foreign_keys_fallback) == str(foreign_keys_odbc)

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_get_foreign_keys(self, use_sql_fallback, engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            foreign_keys = dialect.get_foreign_keys(connection=c, schema=self.schema, table_name="s",
                                                    use_sql_fallback=use_sql_fallback)
            expected = [{'name': 'fk_test',
                         'constrained_columns': ['fid1', 'fid2'],
                         'referred_schema': 'test_get_metadata_functions_schema',
                         'referred_table': 't',
                         'referred_columns': ['pid1', 'pid2']}]

            assert foreign_keys == expected

    @pytest.mark.parametrize("use_sql_fallback", [True, False])
    @pytest.mark.parametrize("engine_name", [ENGINE_NONE_DATABASE, ENGINE_SCHEMA_DATABASE, ENGINE_SCHEMA_2_DATABASE])
    def test_get_foreign_keys_table_name_none(self, use_sql_fallback, engine_name):
        with self.engine_map[engine_name].begin() as c:
            dialect = EXADialect()
            foreign_keys = dialect.get_foreign_keys(connection=c, schema=self.schema, table_name=None,
                                                        use_sql_fallback=use_sql_fallback)
            assert foreign_keys == []



# -*- coding: UTF-8 -*-
import time

import pytest
from sqlalchemy.testing import fixtures, config
from sqlalchemy import MetaData, Table, Column, Integer


table_counts = [1,50,100]
column_counts = [3,30,300]

class LargeMetadataTest(fixtures.TablesTest):
    __backend__ = True


    def create_table_ddl(self, schema, name, column_count):
        table_template = "CREATE TABLE {schema}.{table} ({columns})"
        columns = ["column_column_%s int" % i for i in range(column_count)]
        columns_str = ",".join(columns)
        table_str = table_template.format(schema=schema, table=name, columns=columns_str)
        return table_str

    @classmethod
    def define_tables(cls, metadata):
        cls.schema = "LargeMetadataTest".upper()

    def get_table_name(self, table_count, column_count, i):
        return "table_%s_%s_table_%s" % (table_count, column_count, i)

    def test_reflect_table_object(self):
        for table_count in table_counts:
            for column_count in column_counts:
                with config.db.begin() as c:
                    try:
                        c.execute("DROP SCHEMA %s CASCADE" % self.schema)
                    except Exception as e:
                        print(e)
                        pass
                    c.execute("CREATE SCHEMA %s" % self.schema)
                    for i in range(table_count):
                        table_name=self.get_table_name(table_count,column_count,i)
                        table_ddl=self.create_table_ddl(self.schema,table_name,column_count)
                        c.execute(table_ddl)

                meta = MetaData(bind=config.db)
                table_name = self.get_table_name(table_count, column_count, 0)
                start = time.time()
                users_table = Table(table_name,meta,autoload=True, schema=self.schema)
                end = time.time()
                print("table load timer: attempt: 1, table_count: %s, column_count: %s, time: %s"%(table_count,column_count,
                      (end-start)))
                start = time.time()
                users_table = Table(table_name,meta,autoload=True, schema=self.schema)
                end = time.time()
                print("table load timer: attempt: 2, table_count: %s, column_count: %s, time: %s"%(table_count,column_count,
                      (end-start)))
                meta = None

    def test_reflect_metadata_object(self):
        for table_count in table_counts:
            for column_count in column_counts:
                with config.db.begin() as c:
                    try:
                        c.execute("DROP SCHEMA %s CASCADE" % self.schema)
                    except Exception as e:
                        print(e)
                        pass
                    c.execute("CREATE SCHEMA %s" % self.schema)
                    for i in range(table_count):
                        table_name=self.get_table_name(table_count,column_count,i)
                        table_ddl=self.create_table_ddl(self.schema,table_name,column_count)
                        c.execute(table_ddl)
                meta = MetaData(bind=config.db)
                start = time.time()
                meta.reflect()
                end = time.time()
                print("all tables (MetaData.reflect) load timer: table_count: %s, column_count: %s, time: %s"%(table_count,column_count,
                      (end-start)))



"""
Connect string::

    exa+pyodbc://<username>:<password>@<dsnname>
    exa+pyodbc://<username>:<password>@<ip_range>:<port>/<schema>?<options>

"""

import re
import six
import sys
from sqlalchemy_exasol.base import EXADialect, EXAExecutionContext
from sqlalchemy.connectors.pyodbc import PyODBCConnector
from sqlalchemy.util.langhelpers import asbool
from distutils.version import LooseVersion


class EXADialect_pyodbc(EXADialect, PyODBCConnector):

    execution_ctx_cls = EXAExecutionContext

    driver_version = None
    server_version_info = None

    def __init__(self, **kw):
        super(EXADialect_pyodbc, self).__init__(**kw)

    def get_driver_version(self, connection):
        # LooseVersion will also work with interim versions like '4.2.7dev1' or '5.0.rc4'
        if self.driver_version is None:
            self.driver_version = LooseVersion( connection.connection.getinfo( self.dbapi.SQL_DRIVER_VER ) or '2.0.0' )
        return self.driver_version

    def _get_server_version_info(self, connection):
        if self.server_version_info is None:
            # need to check if current version of EXAODBC returns proper server version
            if self.get_driver_version(connection) >= LooseVersion('4.2.1'):
                # v4.2.1 and above should deliver usable SQL_DBMS_VER
                result = connection.connection.getinfo( self.dbapi.SQL_DBMS_VER ).split('.')
            else:
                # Older versions do not include patchlevels, so we need to get info through SQL call
                query = "select PARAM_VALUE from SYS.EXA_METADATA where PARAM_NAME = 'databaseProductVersion'"
                result = connection.execute(query).fetchone()[0].split('.')

            # last version position can something like: '12-S' for an EXASolo
            self.server_version_info = (int(result[0]), int(result[1]), int(result[2].split('-')[0]))

        # return cached info
        return self.server_version_info

    if sys.platform == 'darwin':
        def connect(self, *cargs, **cparams):
            # Get connection
            conn = super(EXADialect_pyodbc, self).connect(*cargs, **cparams)

            # Set up encodings
            conn.setdecoding(self.dbapi.SQL_CHAR, encoding='utf-8')
            conn.setdecoding(self.dbapi.SQL_WCHAR, encoding='utf-8')
            conn.setdecoding(self.dbapi.SQL_WMETADATA, encoding='utf-8')

            if six.PY2:
                conn.setencoding(str, encoding='utf-8')
                conn.setencoding(unicode, encoding='utf-8')
            else:
                conn.setencoding(encoding='utf-8')

            # Return connection
            return conn

    def create_connect_args(self, url):
        """
        Connection strings are EXASolution specific. See EXASolution manual
        on Connection-String-Parameters
        """
        opts = url.translate_connect_args(username='user')
        opts.update(url.query)
        # always enable efficient conversion to Python types: see https://www.exasol.com/support/browse/EXASOL-898
        opts['INTTYPESINRESULTSIFPOSSIBLE'] = 'y'

        keys = opts
        query = url.query

        connect_args = {}
        for param in ('ansi', 'unicode_results', 'autocommit'):
            if param in keys:
                connect_args[param.upper()] = asbool(keys.pop(param))

        dsn_connection = 'dsn' in keys or \
                        ('host' in keys and 'port' not in keys)
        if dsn_connection:
            connectors = ['DSN=%s' % (keys.pop('dsn', '') or \
                        keys.pop('host', ''))]
        else:
            connectors = ["DRIVER={%s}" %
                    keys.pop('driver', None)]

        port = ''
        if 'port' in keys and not 'port' in query:
            port = ':%d' % int(keys.pop('port'))

        connectors.extend(['EXAHOST=%s%s' % (keys.pop('host', ''), port),
                          'EXASCHEMA=%s' % keys.pop('database', '')])

        user = keys.pop("user", None)
        if user:
            connectors.append("UID=%s" % user)
            connectors.append("PWD=%s" % keys.pop('password', ''))
        else:
            connectors.append("Trusted_Connection=Yes")

        # if set to 'Yes', the ODBC layer will try to automagically
        # convert textual data from your database encoding to your
        # client encoding.  This should obviously be set to 'No' if
        # you query a cp1253 encoded database from a latin1 client...
        if 'odbc_autotranslate' in keys:
            connectors.append("AutoTranslate=%s" %
                                keys.pop("odbc_autotranslate"))

        connectors.extend(['%s=%s' % (k, v) for k, v in sorted(six.iteritems(keys))])
        return [[";".join(connectors)], connect_args]

    def is_disconnect(self, e, connection, cursor):
        if isinstance(e, self.dbapi.Error):
            error_codes = {
                '40004', # Connection lost.
                '40009', # Connection lost after internal server error.
                '40018', # Connection lost after system running out of memory.
                '40020', # Connection lost after system running out of memory.
            }
            exasol_error_codes = {
                'HY000': (  # Generic Exasol error code
                    re.compile(six.u(r'operation timed out'), re.IGNORECASE),
                    re.compile(six.u(r'connection lost'), re.IGNORECASE),
                    re.compile(six.u(r'Socket closed by peer'), re.IGNORECASE),
                )
            }

            error_code, error_msg = e.args[:2]

            # import pdb; pdb.set_trace()
            if error_code in exasol_error_codes:
                # Check exasol error
                for msg_re in exasol_error_codes[error_code]:
                    if msg_re.search(error_msg):
                        return True

                return False

            # Check Pyodbc error
            return error_code in error_codes

        return super(EXADialect_pyodbc, self).is_disconnect(e, connection, cursor)


dialect = EXADialect_pyodbc



from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import Column
from sqlalchemy.sql import crud
from sqlalchemy.sql.base import _generative
from sqlalchemy.sql.expression import ValuesBase, and_, UpdateBase, ColumnClause


class Merge(UpdateBase):

    __visit_name__ = 'merge'

    def __init__(self, target_table, source_expr, on):
        self._target_table = target_table
        self._source_expr = source_expr
        self._on = on
        self._on_columns = []
        elements_to_check = list(on.get_children())
        for e in elements_to_check:
            if isinstance(e, Column):
                if e.table == self._target_table:
                    self._on_columns.append(e)
            else:
                elements_to_check.extend(e.get_children())
        self._merge_update_values = None
        self._update_where = None
        self._merge_insert_values = None
        self._insert_where = None
        self._merge_delete = False
        self._delete_where = None

    def _get_source_cols(self):
        source_cols = {}
        elements_to_check = list(self._source_expr.get_children())
        for e in elements_to_check:
            if isinstance(e, ColumnClause):
                source_cols[e.name] = e
        return source_cols

    @_generative
    def update(self, values=None, where=None):
        # by default update all columns of target table that are
        # present in the source expression but exclude columns part
        # of the MERGE ON condition
        if values is None:
            source_cols = self._get_source_cols()
            values = {}
            for c in self._target_table.c:
                if c not in self._on_columns and c.name in source_cols:
                    values[c] = source_cols[c.name]
        self._merge_update_values = ValuesBase(self._target_table, values, [])
        if where is not None:
            if self._merge_delete:
                self._delete_where = self._append_where(self._delete_where, where)
            else:
                self._update_where = self._append_where(self._update_where, where)

    @_generative
    def insert(self, values=None, where=None):
        # by default set all columns of target table that are
        # present in the source expression but exclude columns part
        # of the MERGE ON condition
        if values is None:
            source_cols = self._get_source_cols()
            values = {}
            for c in self._target_table.c:
                if c.name in source_cols:
                    values[c] = source_cols[c.name]
        self._merge_insert_values = ValuesBase(self._target_table, values, [])
        if where is not None:
            self._insert_where = self._append_where(self._insert_where, where)

    @_generative
    def delete(self, where=None):
        self._merge_delete = True
        if self._merge_update_values is None and where is not None:
            self._delete_where = self._append_where(self._delete_where, where)

    def _append_where(self, where, addition):
        if where is not None:
            where = and_(where, addition)
        else:
            where = addition
        return where


def merge(target_table, source_expr, on):
    return Merge(target_table, source_expr, on)


@compiles(Merge, 'exasol')
def visit_merge(element, compiler, **kw):
    msql = "MERGE INTO %s " % compiler.process(element._target_table,
                                              asfrom=True)
    msql += "USING %s " % compiler.process(element._source_expr,
                                           asfrom=True)
    msql += "ON ( %s ) " % compiler.process(element._on)

    if element._merge_update_values is not None:
        cols = crud._get_crud_params(compiler, element._merge_update_values)
        msql += "\nWHEN MATCHED THEN UPDATE SET "
        msql += ', '.join(compiler.visit_column(c[0]) + '=' + c[1] for c in cols)
        if element._merge_delete:
            msql += "\nDELETE "
            if element._delete_where is not None:
                msql += " WHERE %s" % compiler.process(element._delete_where)
        else:
            if element._update_where is not None:
                msql += " WHERE %s" % compiler.process(element._update_where)
    else:
        if element._merge_delete:
            msql += "\nWHEN MATCHED THEN DELETE "
            if element._delete_where is not None:
                msql += "WHERE %s" % compiler.process(element._delete_where)
    if element._merge_insert_values is not None:
        cols = crud._get_crud_params(compiler, element._merge_insert_values)
        msql += "\nWHEN NOT MATCHED THEN INSERT "
        msql += "(%s) " % ", ".join(compiler.visit_column(c[0]) for c in cols)
        msql += "VALUES (%s) " % ", ".join(c[1] for c in cols)
        if element._insert_where is not None:
            msql += "WHERE %s" % compiler.process(element._insert_where)

    return msql



import unittest
from sqlalchemy import *
from sqlalchemy import testing
from sqlalchemy.dialects import mysql
from sqlalchemy.engine import default
from sqlalchemy.testing import AssertsCompiledSQL, eq_, fixtures
from sqlalchemy.testing.schema import Table, Column


class _UpdateTestBase(object):

    @classmethod
    def define_tables(cls, metadata):
        Table('mytable', metadata,
              Column('myid', Integer),
              Column('name', String(30)),
              Column('description', String(50)))
        Table('myothertable', metadata,
              Column('otherid', Integer),
              Column('othername', String(30)))
        Table('users', metadata,
              Column('id', Integer, primary_key=True,
                     test_needs_autoincrement=True),
              Column('name', String(30), nullable=False))
        Table('addresses', metadata,
              Column('id', Integer, primary_key=True,
                     test_needs_autoincrement=True),
              Column('user_id', None, ForeignKey('users.id')),
              Column('name', String(30), nullable=False),
              Column('email_address', String(50), nullable=False))
        Table('dingalings', metadata,
              Column('id', Integer, primary_key=True,
                     test_needs_autoincrement=True),
              Column('address_id', None, ForeignKey('addresses.id')),
              Column('data', String(30)))

    @classmethod
    def fixtures(cls):
        return dict(
            users=(
                ('id', 'name'),
                (7, 'jack'),
                (8, 'ed'),
                (9, 'fred'),
                (10, 'chuck')
            ),
            addresses = (
                ('id', 'user_id', 'name', 'email_address'),
                (1, 7, 'x', 'jack@bean.com'),
                (2, 8, 'x', 'ed@wood.com'),
                (3, 8, 'x', 'ed@bettyboop.com'),
                (4, 8, 'x', 'ed@lala.com'),
                (5, 9, 'x', 'fred@fred.com')
            ),
            dingalings = (
                ('id', 'address_id', 'data'),
                (1, 2, 'ding 1/2'),
                (2, 5, 'ding 2/5')
            ),
        )


class UpdateTest(_UpdateTestBase, fixtures.TablesTest):
    __backend__ = True

    @unittest.skipIf(testing.db.dialect.driver == 'turbodbc',
                     'not supported by turbodbc')
    def test_update_simple(self):
        """test simple update and assert that exasol returns the right rowcount"""
        users =  self.tables.users
        result = testing.db.execute(users.update().values(name="peter").where(users.c.id == 10))
        expected = [
                (7, 'jack'),
                (8, 'ed'),
                (9, 'fred'),
                (10, 'peter')
                ]
        assert result.rowcount ==  1
        self._assert_users(users, expected)

    @unittest.skipIf(testing.db.dialect.driver == 'turbodbc',
                     'not supported by turbodbc')
    def test_update_simple_multiple_rows_rowcount(self):
        """test simple update and assert that exasol returns the right rowcount"""
        users =  self.tables.users
        result = testing.db.execute(users.update().values(name="peter").where(users.c.id >= 9))
        expected = [
                (7, 'jack'),
                (8, 'ed'),
                (9, 'peter'),
                (10, 'peter')
                ]
        assert result.rowcount == 2 
        self._assert_users(users, expected)

    @unittest.skipIf(testing.db.dialect.driver == 'turbodbc',
                     'not supported by turbodbc')
    def test_update_executemany(self):
        """test that update with executemany work as well, but rowcount
        is undefined for executemany updates"""
        users =  self.tables.users

        stmt = users.update().\
                where(users.c.name == bindparam('oldname')).\
                values(name=bindparam('newname'))

        values = [{'oldname':'jack', 'newname':'jack2'},
                  {'oldname':'fred', 'newname':'fred2'}]

        result = testing.db.execute(stmt, values)
        expected = [ 
                (7, 'jack2'),
                (8, 'ed'),
                (9, 'fred2'),
                (10, 'chuck')
                ]  
        assert result.rowcount ==  -1
        self._assert_users(users, expected)

    def _assert_addresses(self, addresses, expected):
        stmt = addresses.select().order_by(addresses.c.id)
        eq_(testing.db.execute(stmt).fetchall(), expected)

    def _assert_users(self, users, expected):
        stmt = users.select().order_by(users.c.id)
        eq_(testing.db.execute(stmt).fetchall(), expected)



# -*- coding: UTF-8 -*-

import time
from threading import Thread

import pytest
from sqlalchemy import create_engine
from sqlalchemy.testing import fixtures, config
import sqlalchemy.testing as testing
from sqlalchemy_exasol.base import EXADialect


#TODO get_schema_names, get_view_names and get_view_definition didn't cause deadlocks in this scenario
@pytest.mark.skipif("turbodbc" in str(testing.db.url), reason="We currently don't support snapshot metadata requests for turbodbc")
class MetadataTest(fixtures.TablesTest):
    __backend__ = True

    CONNECTION_ECHO = False
    WATCHDOG_ECHO = False

    def create_transaction(self, url, con_name):
        engine = create_engine(config.db.url, echo=self.CONNECTION_ECHO, logging_name="engine" + con_name)
        session = engine.connect().execution_options(autocommit=False)
        return engine, session

    def test_no_deadlock_for_get_table_names_without_fallback(self):
        def without_fallback(session2, schema, table):
            dialect = EXADialect()
            dialect.get_table_names(session2, schema=schema, use_sql_fallback=False)

        self.run_deadlock_for_table(without_fallback)

    def test_deadlock_for_get_table_names_with_fallback(self):
        def with_fallback(session2, schema, table):
            dialect = EXADialect()
            dialect.get_table_names(session2, schema=schema, use_sql_fallback=True)

        with pytest.raises(Exception):
            self.run_deadlock_for_table(with_fallback)
    
    def test_no_deadlock_for_get_columns_without_fallback(self):
        def without_fallback(session2, schema, table):
            dialect = EXADialect()
            dialect.get_columns(session2, schema=schema, table_name=table, use_sql_fallback=False)

        self.run_deadlock_for_table(without_fallback)

    def test_no_deadlock_for_get_columns_with_fallback(self):
        # TODO: Doesnt produce a deadlock anymore since last commit?
        def with_fallback(session2, schema, table):
            dialect = EXADialect()
            dialect.get_columns(session2, schema=schema, table_name=table, use_sql_fallback=True)

        self.run_deadlock_for_table(with_fallback)

    def test_no_deadlock_for_get_pk_constraint_without_fallback(self):
        def without_fallback(session2, schema, table):
            dialect = EXADialect()
            dialect.get_pk_constraint(session2, table_name=table, schema=schema, use_sql_fallback=False)

        self.run_deadlock_for_table(without_fallback)

    def test_no_deadlock_for_get_pk_constraint_with_fallback(self):
        def with_fallback(session2, schema, table):
            dialect = EXADialect()
            dialect.get_pk_constraint(session2, table_name=table, schema=schema, use_sql_fallback=True)

        self.run_deadlock_for_table(with_fallback)

    def test_no_deadlock_for_get_foreign_keys_without_fallback(self):
        def without_fallback(session2, schema, table):
            dialect = EXADialect()
            dialect.get_foreign_keys(session2, table_name=table, schema=schema, use_sql_fallback=False)

        self.run_deadlock_for_table(without_fallback)


    def test_no_deadlock_for_get_foreign_keys_with_fallback(self):
        def with_fallback(session2, schema, table):
            dialect = EXADialect()
            dialect.get_foreign_keys(session2, table_name=table, schema=schema, use_sql_fallback=True)

        self.run_deadlock_for_table(with_fallback)

    def test_no_deadlock_for_get_view_names_without_fallback(self):
        # TODO: think of other scenarios where metadata deadlocks with view could happen
        def without_fallback(session2, schema, table):
            dialect = EXADialect()
            dialect.get_view_names(session2, table_name=table, schema=schema, use_sql_fallback=False)

        self.run_deadlock_for_table(without_fallback)

    def test_no_deadlock_for_get_view_names_with_fallback(self):
        # TODO: think of other scenarios where metadata deadlocks with view could happen
        def with_fallback(session2, schema, table):
            dialect = EXADialect()
            dialect.get_view_names(session2, table_name=table, schema=schema, use_sql_fallback=True)

        self.run_deadlock_for_table(with_fallback)

    def watchdog(self, session0, schema):
        while self.watchdog_run:
            rs = session0.execute("SELECT * FROM SYS.EXA_ALL_SESSIONS")
            rs = [row for row in rs]
            if self.WATCHDOG_ECHO:
                print()
                print("===========================================")
                print("Watchdog")
                print("===========================================")
            for row in rs:
                if self.WATCHDOG_ECHO:
                    print(row)
                if row[7] is not None and "Waiting for" in row[7]:
                    if self.WATCHDOG_ECHO:
                        print("Killing session: %s" % row[0])
                    session0.execute("kill session %s" % row[0])
            if self.WATCHDOG_ECHO:
                print("===========================================")
                print()
            time.sleep(10) # Only change with care, lower values might make tests unreliable

    def run_deadlock_for_table(self, function):
        c1 = config.db.connect()
        url = config.db.url

        schema = "deadlock_get_table_names_test_schema"
        engine0, session0 = self.create_transaction(url, "transaction0")
        try:
            session0.execute("DROP SCHEMA %s cascade" % schema)
        except:
            pass
        session0.execute("CREATE SCHEMA %s" % schema)
        session0.execute("CREATE OR REPLACE TABLE %s.deadlock_test1 (id int PRIMARY KEY)" % schema)
        session0.execute(
            "CREATE OR REPLACE TABLE %s.deadlock_test2 (id int PRIMARY KEY, fk int REFERENCES %s.deadlock_test1(id))" % (
                schema, schema))
        session0.execute("INSERT INTO %s.deadlock_test1 VALUES 1" % schema)
        session0.execute("INSERT INTO %s.deadlock_test2 VALUES (1,1)" % schema)
        session0.execute("commit")
        self.watchdog_run = True
        t1 = Thread(target=self.watchdog, args=(session0, schema))
        t1.start()
        try:
            engine1, session1 = self.create_transaction(url, "transaction1")
            session1.execute("SELECT 1")

            session1.execute("SELECT * FROM %s.deadlock_test2" % schema)
            session1.execute("INSERT INTO %s.deadlock_test1 VALUES 2" % schema)

            engine3, session3 = self.create_transaction(url, "transaction3")
            session3.execute("SELECT 1")
            session3.execute("DELETE FROM %s.deadlock_test2 WHERE false" % schema)
            session3.execute("commit")

            engine2, session2 = self.create_transaction(url, "transaction2")
            session2.execute("SELECT 1")
            function(session2, schema, "deadlock_test2")

            session2.execute("commit")
            session1.execute("commit")
        except Exception as e:
            self.watchdog_run = False
            t1.join()
            raise e
        self.watchdog_run = False
        t1.join()

    def run_deadlock_for_get_view_names(self, function):
        c1 = config.db.connect()
        url = config.db.url

        schema = "deadlock_get_view_names_test_schema"
        engine0, session0 = self.create_transaction(url, "transaction0")
        try:
            session0.execute("DROP SCHEMA %s cascade" % schema)
        except:
            pass
        session0.execute("CREATE SCHEMA %s" % schema)
        session0.execute("CREATE OR REPLACE TABLE %s.deadlock_test_table (id int)" % schema)
        session0.execute(
            "CREATE OR REPLACE VIEW %s.deadlock_test_view_1 AS SELECT * FROM %s.deadlock_test_table" % (schema, schema))
        session0.execute("commit")
        self.watchdog_run = True
        t1 = Thread(target=self.watchdog, args=(session0, schema))
        t1.start()
        try:
            engine1, session1 = self.create_transaction(url, "transaction1")
            session1.execute("SELECT 1")

            session1.execute("SELECT * FROM %s.deadlock_test_view_1" % schema)
            session1.execute(
                "CREATE OR REPLACE VIEW %s.deadlock_test_view_2 AS SELECT * FROM %s.deadlock_test_table" % (
                    schema, schema))

            engine3, session3 = self.create_transaction(url, "transaction3")
            session3.execute("SELECT 1")
            session3.execute("DROP VIEW %s.deadlock_test_view_1" % schema)
            session3.execute("commit")

            engine2, session2 = self.create_transaction(url, "transaction2")
            session2.execute("SELECT 1")
            function(session2, schema)

            session2.execute("commit")
            session1.execute("commit")
        except Exception as e:
            self.watchdog_run = False
            t1.join()
            raise e
        self.watchdog_run = False
        t1.join()




# Version: 0.10

"""
## Installation

First, decide on values for the following configuration variables:

* `versionfile_source`:

  A project-relative pathname into which the generated version strings should
  be written. This is usually a `_version.py` next to your project's main
  `__init__.py` file. If your project uses `src/myproject/__init__.py`, this
  should be `src/myproject/_version.py`. This file should be checked in to
  your VCS as usual: the copy created below by `setup.py versioneer` will
  include code that parses expanded VCS keywords in generated tarballs. The
  'build' and 'sdist' commands will replace it with a copy that has just the
  calculated version string.

*  `versionfile_build`:

  Like `versionfile_source`, but relative to the build directory instead of
  the source directory. These will differ when your setup.py uses
  'package_dir='. If you have `package_dir={'myproject': 'src/myproject'}`,
  then you will probably have `versionfile_build='myproject/_version.py'` and
  `versionfile_source='src/myproject/_version.py'`.

* `tag_prefix`:

  a string, like 'PROJECTNAME-', which appears at the start of all VCS tags.
  If your tags look like 'myproject-1.2.0', then you should use
  tag_prefix='myproject-'. If you use unprefixed tags like '1.2.0', this
  should be an empty string.

* `parentdir_prefix`:

  a string, frequently the same as tag_prefix, which appears at the start of
  all unpacked tarball filenames. If your tarball unpacks into
  'myproject-1.2.0', this should be 'myproject-'.

This tool provides one script, named `versioneer-installer`. That script does
one thing: write a copy of `versioneer.py` into the current directory.

To versioneer-enable your project:

* 1: Run `versioneer-installer` to copy `versioneer.py` into the top of your
  source tree.

* 2: add the following lines to the top of your `setup.py`, with the
  configuration values you decided earlier:

        import versioneer
        versioneer.versionfile_source = 'src/myproject/_version.py'
        versioneer.versionfile_build = 'myproject/_version.py'
        versioneer.tag_prefix = '' # tags are like 1.2.0
        versioneer.parentdir_prefix = 'myproject-' # dirname like 'myproject-1.2.0'

* 3: add the following arguments to the setup() call in your setup.py:

        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),

* 4: now run `setup.py versioneer`, which will create `_version.py`, and
  will modify your `__init__.py` to define `__version__` (by calling a
  function from `_version.py`). It will also modify your `MANIFEST.in` to
  include both `versioneer.py` and the generated `_version.py` in sdist
  tarballs.

* 5: commit these changes to your VCS. To make sure you won't forget,
  `setup.py versioneer` will mark everything it touched for addition.

## Post-Installation Usage

Once established, all uses of your tree from a VCS checkout should get the
current version string. All generated tarballs should include an embedded
version string (so users who unpack them will not need a VCS tool installed).

If you distribute your project through PyPI, then the release process should
boil down to two steps:

* 1: git tag 1.0
* 2: python setup.py register sdist upload

If you distribute it through github (i.e. users use github to generate
tarballs with `git archive`), the process is:

* 1: git tag 1.0
* 2: git push; git push --tags

Currently, all version strings must be based upon a tag. Versioneer will
report "unknown" until your tree has at least one tag in its history. This
restriction will be fixed eventually (see issue #12).

## Version-String Flavors

Code which uses Versioneer can learn about its version string at runtime by
importing `_version` from your main `__init__.py` file and running the
`get_versions()` function. From the "outside" (e.g. in `setup.py`), you can
import the top-level `versioneer.py` and run `get_versions()`.

Both functions return a dictionary with different keys for different flavors
of the version string:

* `['version']`: condensed tag+distance+shortid+dirty identifier. For git,
  this uses the output of `git describe --tags --dirty --always` but strips
  the tag_prefix. For example "0.11-2-g1076c97-dirty" indicates that the tree
  is like the "1076c97" commit but has uncommitted changes ("-dirty"), and
  that this commit is two revisions ("-2-") beyond the "0.11" tag. For
  released software (exactly equal to a known tag), the identifier will only
  contain the stripped tag, e.g. "0.11".

* `['full']`: detailed revision identifier. For Git, this is the full SHA1
  commit id, followed by "-dirty" if the tree contains uncommitted changes,
  e.g. "1076c978a8d3cfc70f408fe5974aa6c092c949ac-dirty".

Some variants are more useful than others. Including `full` in a bug report
should allow developers to reconstruct the exact code being tested (or
indicate the presence of local changes that should be shared with the
developers). `version` is suitable for display in an "about" box or a CLI
`--version` output: it can be easily compared against release notes and lists
of bugs fixed in various releases.

In the future, this will also include a
[PEP-0440](http://legacy.python.org/dev/peps/pep-0440/) -compatible flavor
(e.g. `1.2.post0.dev123`). This loses a lot of information (and has no room
for a hash-based revision id), but is safe to use in a `setup.py`
"`version=`" argument. It also enables tools like *pip* to compare version
strings and evaluate compatibility constraint declarations.

The `setup.py versioneer` command adds the following text to your
`__init__.py` to place a basic version in `YOURPROJECT.__version__`:

    from ._version import get_versions
    __version = get_versions()['version']
    del get_versions

## Updating Versioneer

To upgrade your project to a new release of Versioneer, do the following:

* install the new Versioneer (`pip install -U versioneer` or equivalent)
* re-run `versioneer-installer` in your source tree to replace `versioneer.py`
* edit `setup.py`, if necessary, to include any new configuration settings indicated by the release notes
* re-run `setup.py versioneer` to replace `SRC/_version.py`
* commit any changed files

## Future Directions

This tool is designed to make it easily extended to other version-control
systems: all VCS-specific components are in separate directories like
src/git/ . The top-level `versioneer.py` script is assembled from these
components by running make-versioneer.py . In the future, make-versioneer.py
will take a VCS name as an argument, and will construct a version of
`versioneer.py` that is specific to the given VCS. It might also take the
configuration arguments that are currently provided manually during
installation by editing setup.py . Alternatively, it might go the other
direction and include code from all supported VCS systems, reducing the
number of intermediate scripts.


## License

To make Versioneer easier to embed, all its code is hereby released into the
public domain. The `_version.py` that it creates is also in the public
domain.

"""

import os, sys, re
from distutils.core import Command
from distutils.command.sdist import sdist as _sdist
from distutils.command.build import build as _build

versionfile_source = None
versionfile_build = None
tag_prefix = None
parentdir_prefix = None

VCS = "git"


LONG_VERSION_PY = '''
# This file helps to compute a version number in source trees obtained from
# git-archive tarball (such as those provided by githubs download-from-tag
# feature). Distribution tarballs (build by setup.py sdist) and build
# directories (produced by setup.py build) will contain a much shorter file
# that just contains the computed version number.

# This file is released into the public domain. Generated by
# versioneer-0.10 (https://github.com/warner/python-versioneer)

# these strings will be replaced by git during git-archive
git_refnames = "%(DOLLAR)sFormat:%%d%(DOLLAR)s"
git_full = "%(DOLLAR)sFormat:%%H%(DOLLAR)s"


import subprocess
import sys
import errno


def run_command(commands, args, cwd=None, verbose=False, hide_stderr=False):
    assert isinstance(commands, list)
    p = None
    for c in commands:
        try:
            # remember shell=False, so use git.cmd on windows, not just git
            p = subprocess.Popen([c] + args, cwd=cwd, stdout=subprocess.PIPE,
                                 stderr=(subprocess.PIPE if hide_stderr
                                         else None))
            break
        except EnvironmentError:
            e = sys.exc_info()[1]
            if e.errno == errno.ENOENT:
                continue
            if verbose:
                print("unable to run %%s" %% args[0])
                print(e)
            return None
    else:
        if verbose:
            print("unable to find command, tried %%s" %% (commands,))
        return None
    stdout = p.communicate()[0].strip()
    if sys.version >= '3':
        stdout = stdout.decode()
    if p.returncode != 0:
        if verbose:
            print("unable to run %%s (error)" %% args[0])
        return None
    return stdout


import sys
import re
import os.path

def get_expanded_variables(versionfile_abs):
    # the code embedded in _version.py can just fetch the value of these
    # variables. When used from setup.py, we don't want to import
    # _version.py, so we do it with a regexp instead. This function is not
    # used from _version.py.
    variables = {}
    try:
        f = open(versionfile_abs,"r")
        for line in f.readlines():
            if line.strip().startswith("git_refnames ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    variables["refnames"] = mo.group(1)
            if line.strip().startswith("git_full ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    variables["full"] = mo.group(1)
        f.close()
    except EnvironmentError:
        pass
    return variables

def versions_from_expanded_variables(variables, tag_prefix, verbose=False):
    refnames = variables["refnames"].strip()
    if refnames.startswith("$Format"):
        if verbose:
            print("variables are unexpanded, not using")
        return {} # unexpanded, so not in an unpacked git-archive tarball
    refs = set([r.strip() for r in refnames.strip("()").split(",")])
    # starting in git-1.8.3, tags are listed as "tag: foo-1.0" instead of
    # just "foo-1.0". If we see a "tag: " prefix, prefer those.
    TAG = "tag: "
    tags = set([r[len(TAG):] for r in refs if r.startswith(TAG)])
    if not tags:
        # Either we're using git < 1.8.3, or there really are no tags. We use
        # a heuristic: assume all version tags have a digit. The old git %%d
        # expansion behaves like git log --decorate=short and strips out the
        # refs/heads/ and refs/tags/ prefixes that would let us distinguish
        # between branches and tags. By ignoring refnames without digits, we
        # filter out many common branch names like "release" and
        # "stabilization", as well as "HEAD" and "master".
        tags = set([r for r in refs if re.search(r'\d', r)])
        if verbose:
            print("discarding '%%s', no digits" %% ",".join(refs-tags))
    if verbose:
        print("likely tags: %%s" %% ",".join(sorted(tags)))
    for ref in sorted(tags):
        # sorting will prefer e.g. "2.0" over "2.0rc1"
        if ref.startswith(tag_prefix):
            r = ref[len(tag_prefix):]
            if verbose:
                print("picking %%s" %% r)
            return { "version": r,
                     "full": variables["full"].strip() }
    # no suitable tags, so we use the full revision id
    if verbose:
        print("no suitable tags, using full revision id")
    return { "version": variables["full"].strip(),
             "full": variables["full"].strip() }

def versions_from_vcs(tag_prefix, root, verbose=False):
    # this runs 'git' from the root of the source tree. This only gets called
    # if the git-archive 'subst' variables were *not* expanded, and
    # _version.py hasn't already been rewritten with a short version string,
    # meaning we're inside a checked out source tree.

    if not os.path.exists(os.path.join(root, ".git")):
        if verbose:
            print("no .git in %%s" %% root)
        return {}

    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]
    stdout = run_command(GITS, ["describe", "--tags", "--dirty", "--always"],
                         cwd=root)
    if stdout is None:
        return {}
    if not stdout.startswith(tag_prefix):
        if verbose:
            print("tag '%%s' doesn't start with prefix '%%s'" %% (stdout, tag_prefix))
        return {}
    tag = stdout[len(tag_prefix):]
    stdout = run_command(GITS, ["rev-parse", "HEAD"], cwd=root)
    if stdout is None:
        return {}
    full = stdout.strip()
    if tag.endswith("-dirty"):
        full += "-dirty"
    return {"version": tag, "full": full}


def versions_from_parentdir(parentdir_prefix, root, verbose=False):
    # Source tarballs conventionally unpack into a directory that includes
    # both the project name and a version string.
    dirname = os.path.basename(root)
    if not dirname.startswith(parentdir_prefix):
        if verbose:
            print("guessing rootdir is '%%s', but '%%s' doesn't start with prefix '%%s'" %%
                  (root, dirname, parentdir_prefix))
        return None
    return {"version": dirname[len(parentdir_prefix):], "full": ""}

tag_prefix = "%(TAG_PREFIX)s"
parentdir_prefix = "%(PARENTDIR_PREFIX)s"
versionfile_source = "%(VERSIONFILE_SOURCE)s"

def get_versions(default={"version": "unknown", "full": ""}, verbose=False):
    # I am in _version.py, which lives at ROOT/VERSIONFILE_SOURCE. If we have
    # __file__, we can work backwards from there to the root. Some
    # py2exe/bbfreeze/non-CPython implementations don't do __file__, in which
    # case we can only use expanded variables.

    variables = { "refnames": git_refnames, "full": git_full }
    ver = versions_from_expanded_variables(variables, tag_prefix, verbose)
    if ver:
        return ver

    try:
        root = os.path.abspath(__file__)
        # versionfile_source is the relative path from the top of the source
        # tree (where the .git directory might live) to this file. Invert
        # this to find the root from __file__.
        for i in range(len(versionfile_source.split("/"))):
            root = os.path.dirname(root)
    except NameError:
        return default

    return (versions_from_vcs(tag_prefix, root, verbose)
            or versions_from_parentdir(parentdir_prefix, root, verbose)
            or default)

'''


import subprocess
import sys
import errno


def run_command(commands, args, cwd=None, verbose=False, hide_stderr=False):
    assert isinstance(commands, list)
    p = None
    for c in commands:
        try:
            # remember shell=False, so use git.cmd on windows, not just git
            p = subprocess.Popen([c] + args, cwd=cwd, stdout=subprocess.PIPE,
                                 stderr=(subprocess.PIPE if hide_stderr
                                         else None))
            break
        except EnvironmentError:
            e = sys.exc_info()[1]
            if e.errno == errno.ENOENT:
                continue
            if verbose:
                print("unable to run %s" % args[0])
                print(e)
            return None
    else:
        if verbose:
            print("unable to find command, tried %s" % (commands,))
        return None
    stdout = p.communicate()[0].strip()
    if sys.version >= '3':
        stdout = stdout.decode()
    if p.returncode != 0:
        if verbose:
            print("unable to run %s (error)" % args[0])
        return None
    return stdout


import sys
import re
import os.path

def get_expanded_variables(versionfile_abs):
    # the code embedded in _version.py can just fetch the value of these
    # variables. When used from setup.py, we don't want to import
    # _version.py, so we do it with a regexp instead. This function is not
    # used from _version.py.
    variables = {}
    try:
        f = open(versionfile_abs,"r")
        for line in f.readlines():
            if line.strip().startswith("git_refnames ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    variables["refnames"] = mo.group(1)
            if line.strip().startswith("git_full ="):
                mo = re.search(r'=\s*"(.*)"', line)
                if mo:
                    variables["full"] = mo.group(1)
        f.close()
    except EnvironmentError:
        pass
    return variables

def versions_from_expanded_variables(variables, tag_prefix, verbose=False):
    refnames = variables["refnames"].strip()
    if refnames.startswith("$Format"):
        if verbose:
            print("variables are unexpanded, not using")
        return {} # unexpanded, so not in an unpacked git-archive tarball
    refs = set([r.strip() for r in refnames.strip("()").split(",")])
    # starting in git-1.8.3, tags are listed as "tag: foo-1.0" instead of
    # just "foo-1.0". If we see a "tag: " prefix, prefer those.
    TAG = "tag: "
    tags = set([r[len(TAG):] for r in refs if r.startswith(TAG)])
    if not tags:
        # Either we're using git < 1.8.3, or there really are no tags. We use
        # a heuristic: assume all version tags have a digit. The old git %d
        # expansion behaves like git log --decorate=short and strips out the
        # refs/heads/ and refs/tags/ prefixes that would let us distinguish
        # between branches and tags. By ignoring refnames without digits, we
        # filter out many common branch names like "release" and
        # "stabilization", as well as "HEAD" and "master".
        tags = set([r for r in refs if re.search(r'\d', r)])
        if verbose:
            print("discarding '%s', no digits" % ",".join(refs-tags))
    if verbose:
        print("likely tags: %s" % ",".join(sorted(tags)))
    for ref in sorted(tags):
        # sorting will prefer e.g. "2.0" over "2.0rc1"
        if ref.startswith(tag_prefix):
            r = ref[len(tag_prefix):]
            if verbose:
                print("picking %s" % r)
            return { "version": r,
                     "full": variables["full"].strip() }
    # no suitable tags, so we use the full revision id
    if verbose:
        print("no suitable tags, using full revision id")
    return { "version": variables["full"].strip(),
             "full": variables["full"].strip() }

def versions_from_vcs(tag_prefix, root, verbose=False):
    # this runs 'git' from the root of the source tree. This only gets called
    # if the git-archive 'subst' variables were *not* expanded, and
    # _version.py hasn't already been rewritten with a short version string,
    # meaning we're inside a checked out source tree.

    if not os.path.exists(os.path.join(root, ".git")):
        if verbose:
            print("no .git in %s" % root)
        return {}

    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]
    stdout = run_command(GITS, ["describe", "--tags", "--dirty", "--always"],
                         cwd=root)
    if stdout is None:
        return {}
    if not stdout.startswith(tag_prefix):
        if verbose:
            print("tag '%s' doesn't start with prefix '%s'" % (stdout, tag_prefix))
        return {}
    tag = stdout[len(tag_prefix):]
    stdout = run_command(GITS, ["rev-parse", "HEAD"], cwd=root)
    if stdout is None:
        return {}
    full = stdout.strip()
    if tag.endswith("-dirty"):
        full += "-dirty"
    return {"version": tag, "full": full}


def versions_from_parentdir(parentdir_prefix, root, verbose=False):
    # Source tarballs conventionally unpack into a directory that includes
    # both the project name and a version string.
    dirname = os.path.basename(root)
    if not dirname.startswith(parentdir_prefix):
        if verbose:
            print("guessing rootdir is '%s', but '%s' doesn't start with prefix '%s'" %
                  (root, dirname, parentdir_prefix))
        return None
    return {"version": dirname[len(parentdir_prefix):], "full": ""}
import os.path
import sys

# os.path.relpath only appeared in Python-2.6 . Define it here for 2.5.
def os_path_relpath(path, start=os.path.curdir):
    """Return a relative version of a path"""

    if not path:
        raise ValueError("no path specified")

    start_list = [x for x in os.path.abspath(start).split(os.path.sep) if x]
    path_list = [x for x in os.path.abspath(path).split(os.path.sep) if x]

    # Work out how much of the filepath is shared by start and path.
    i = len(os.path.commonprefix([start_list, path_list]))

    rel_list = [os.path.pardir] * (len(start_list)-i) + path_list[i:]
    if not rel_list:
        return os.path.curdir
    return os.path.join(*rel_list)

def do_vcs_install(manifest_in, versionfile_source, ipy):
    GITS = ["git"]
    if sys.platform == "win32":
        GITS = ["git.cmd", "git.exe"]
    files = [manifest_in, versionfile_source, ipy]
    try:
        me = __file__
        if me.endswith(".pyc") or me.endswith(".pyo"):
            me = os.path.splitext(me)[0] + ".py"
        versioneer_file = os_path_relpath(me)
    except NameError:
        versioneer_file = "versioneer.py"
    files.append(versioneer_file)
    present = False
    try:
        f = open(".gitattributes", "r")
        for line in f.readlines():
            if line.strip().startswith(versionfile_source):
                if "export-subst" in line.strip().split()[1:]:
                    present = True
        f.close()
    except EnvironmentError:
        pass    
    if not present:
        f = open(".gitattributes", "a+")
        f.write("%s export-subst\n" % versionfile_source)
        f.close()
        files.append(".gitattributes")
    run_command(GITS, ["add", "--"] + files)

SHORT_VERSION_PY = """
# This file was generated by 'versioneer.py' (0.10) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '%(version)s'
version_full = '%(full)s'
def get_versions(default={}, verbose=False):
    return {'version': version_version, 'full': version_full}

"""

DEFAULT = {"version": "unknown", "full": "unknown"}

def versions_from_file(filename):
    versions = {}
    try:
        f = open(filename)
    except EnvironmentError:
        return versions
    for line in f.readlines():
        mo = re.match("version_version = '([^']+)'", line)
        if mo:
            versions["version"] = mo.group(1)
        mo = re.match("version_full = '([^']+)'", line)
        if mo:
            versions["full"] = mo.group(1)
    f.close()
    return versions

def write_to_version_file(filename, versions):
    f = open(filename, "w")
    f.write(SHORT_VERSION_PY % versions)
    f.close()
    print("set %s to '%s'" % (filename, versions["version"]))

def get_root():
    try:
        return os.path.dirname(os.path.abspath(__file__))
    except NameError:
        return os.path.dirname(os.path.abspath(sys.argv[0]))

def get_versions(default=DEFAULT, verbose=False):
    # returns dict with two keys: 'version' and 'full'
    assert versionfile_source is not None, "please set versioneer.versionfile_source"
    assert tag_prefix is not None, "please set versioneer.tag_prefix"
    assert parentdir_prefix is not None, "please set versioneer.parentdir_prefix"
    # I am in versioneer.py, which must live at the top of the source tree,
    # which we use to compute the root directory. py2exe/bbfreeze/non-CPython
    # don't have __file__, in which case we fall back to sys.argv[0] (which
    # ought to be the setup.py script). We prefer __file__ since that's more
    # robust in cases where setup.py was invoked in some weird way (e.g. pip)
    root = get_root()
    versionfile_abs = os.path.join(root, versionfile_source)

    # extract version from first of _version.py, 'git describe', parentdir.
    # This is meant to work for developers using a source checkout, for users
    # of a tarball created by 'setup.py sdist', and for users of a
    # tarball/zipball created by 'git archive' or github's download-from-tag
    # feature.

    variables = get_expanded_variables(versionfile_abs)
    if variables:
        ver = versions_from_expanded_variables(variables, tag_prefix)
        if ver:
            if verbose: print("got version from expanded variable %s" % ver)
            return ver

    ver = versions_from_file(versionfile_abs)
    if ver:
        if verbose: print("got version from file %s %s" % (versionfile_abs,ver))
        return ver

    ver = versions_from_vcs(tag_prefix, root, verbose)
    if ver:
        if verbose: print("got version from git %s" % ver)
        return ver

    ver = versions_from_parentdir(parentdir_prefix, root, verbose)
    if ver:
        if verbose: print("got version from parentdir %s" % ver)
        return ver

    if verbose: print("got version from default %s" % ver)
    return default

def get_version(verbose=False):
    return get_versions(verbose=verbose)["version"]

class cmd_version(Command):
    description = "report generated version string"
    user_options = []
    boolean_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        ver = get_version(verbose=True)
        print("Version is currently: %s" % ver)


class cmd_build(_build):
    def run(self):
        versions = get_versions(verbose=True)
        _build.run(self)
        # now locate _version.py in the new build/ directory and replace it
        # with an updated value
        target_versionfile = os.path.join(self.build_lib, versionfile_build)
        print("UPDATING %s" % target_versionfile)
        os.unlink(target_versionfile)
        f = open(target_versionfile, "w")
        f.write(SHORT_VERSION_PY % versions)
        f.close()

if 'cx_Freeze' in sys.modules:  # cx_freeze enabled?
    from cx_Freeze.dist import build_exe as _build_exe

    class cmd_build_exe(_build_exe):
        def run(self):
            versions = get_versions(verbose=True)
            target_versionfile = versionfile_source
            print("UPDATING %s" % target_versionfile)
            os.unlink(target_versionfile)
            f = open(target_versionfile, "w")
            f.write(SHORT_VERSION_PY % versions)
            f.close()
            _build_exe.run(self)
            os.unlink(target_versionfile)
            f = open(versionfile_source, "w")
            f.write(LONG_VERSION_PY % {"DOLLAR": "$",
                                       "TAG_PREFIX": tag_prefix,
                                       "PARENTDIR_PREFIX": parentdir_prefix,
                                       "VERSIONFILE_SOURCE": versionfile_source,
                                       })
            f.close()

class cmd_sdist(_sdist):
    def run(self):
        versions = get_versions(verbose=True)
        self._versioneer_generated_versions = versions
        # unless we update this, the command will keep using the old version
        self.distribution.metadata.version = versions["version"]
        return _sdist.run(self)

    def make_release_tree(self, base_dir, files):
        _sdist.make_release_tree(self, base_dir, files)
        # now locate _version.py in the new base_dir directory (remembering
        # that it may be a hardlink) and replace it with an updated value
        target_versionfile = os.path.join(base_dir, versionfile_source)
        print("UPDATING %s" % target_versionfile)
        os.unlink(target_versionfile)
        f = open(target_versionfile, "w")
        f.write(SHORT_VERSION_PY % self._versioneer_generated_versions)
        f.close()

INIT_PY_SNIPPET = """
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
"""

class cmd_update_files(Command):
    description = "install/upgrade Versioneer files: __init__.py SRC/_version.py"
    user_options = []
    boolean_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        print(" creating %s" % versionfile_source)
        f = open(versionfile_source, "w")
        f.write(LONG_VERSION_PY % {"DOLLAR": "$",
                                   "TAG_PREFIX": tag_prefix,
                                   "PARENTDIR_PREFIX": parentdir_prefix,
                                   "VERSIONFILE_SOURCE": versionfile_source,
                                   })
        f.close()

        ipy = os.path.join(os.path.dirname(versionfile_source), "__init__.py")
        try:
            old = open(ipy, "r").read()
        except EnvironmentError:
            old = ""
        if INIT_PY_SNIPPET not in old:
            print(" appending to %s" % ipy)
            f = open(ipy, "a")
            f.write(INIT_PY_SNIPPET)
            f.close()
        else:
            print(" %s unmodified" % ipy)

        # Make sure both the top-level "versioneer.py" and versionfile_source
        # (PKG/_version.py, used by runtime code) are in MANIFEST.in, so
        # they'll be copied into source distributions. Pip won't be able to
        # install the package without this.
        manifest_in = os.path.join(get_root(), "MANIFEST.in")
        simple_includes = set()
        try:
            for line in open(manifest_in, "r").readlines():
                if line.startswith("include "):
                    for include in line.split()[1:]:
                        simple_includes.add(include)
        except EnvironmentError:
            pass
        # That doesn't cover everything MANIFEST.in can do
        # (http://docs.python.org/2/distutils/sourcedist.html#commands), so
        # it might give some false negatives. Appending redundant 'include'
        # lines is safe, though.
        if "versioneer.py" not in simple_includes:
            print(" appending 'versioneer.py' to MANIFEST.in")
            f = open(manifest_in, "a")
            f.write("include versioneer.py\n")
            f.close()
        else:
            print(" 'versioneer.py' already in MANIFEST.in")
        if versionfile_source not in simple_includes:
            print(" appending versionfile_source ('%s') to MANIFEST.in" %
                  versionfile_source)
            f = open(manifest_in, "a")
            f.write("include %s\n" % versionfile_source)
            f.close()
        else:
            print(" versionfile_source already in MANIFEST.in")

        # Make VCS-specific changes. For git, this means creating/changing
        # .gitattributes to mark _version.py for export-time keyword
        # substitution.
        do_vcs_install(manifest_in, versionfile_source, ipy)

def get_cmdclass():
    cmds = {'version': cmd_version,
            'versioneer': cmd_update_files,
            'build': cmd_build,
            'sdist': cmd_sdist,
            }
    if 'cx_Freeze' in sys.modules:  # cx_freeze enabled?
        cmds['build_exe'] = cmd_build_exe
        del cmds['build']

    return cmds



from sqlalchemy.dialects import registry

registry.register("exa.pyodbc", "sqlalchemy_exasol.pyodbc", "EXADialect_pyodbc")
registry.register("exa.turbodbc", "sqlalchemy_exasol.turbodbc", "EXADialect_turbodbc")

from sqlalchemy.testing.plugin.pytestplugin import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Defines the setup for this package.

Do not forget to document your changes in CHANGES.md

:file: setup.py
:authors: Blue Yonder GmbH
:date: 2014/02/11
"""
from setuptools import setup

import os


## Get long_description from README.md:
here = os.path.dirname(os.path.abspath(__file__))
long_description = ''
with open(os.path.join(here, 'README.rst')) as f:
    long_description = f.read().strip()

setup(
    name='sqlalchemy_exasol',
    setup_requires=["setuptools_scm"],
    license='License :: OSI Approved :: BSD License',
    url='https://github.com/blue-yonder/sqlalchemy_exasol',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Database'
    ],
    description='EXASOL dialect for SQLAlchemy',
    long_description=long_description,
    author='Blue Yonder GmbH',
    packages=['sqlalchemy_exasol'],
    install_requires=["SQLAlchemy>=1.0.4, <2", "pyodbc>=3.0.6", "six>=1.5"],
    extras_require={'turbodbc': ['turbodbc>=1.1.0']},
    tests_require=['pytest', 'mock>=1.0.1'],
    test_suite='pytest.main',
    use_scm_version={"write_to": "sqlalchemy_exasol/_version.py"},
    entry_points={
          'sqlalchemy.dialects': ['exa.pyodbc = sqlalchemy_exasol.pyodbc:EXADialect_pyodbc',
                                  'exa.turbodbc = sqlalchemy_exasol.turbodbc:EXADialect_turbodbc']
    },
)



from sqlalchemy import testing
from sqlalchemy.engine import url as sa_url
from sqlalchemy.testing import fixtures
from sqlalchemy.testing import eq_

from sqlalchemy_exasol.turbodbc import (EXADialect_turbodbc, DEFAULT_CONNECTION_PARAMS,
                                        DEFAULT_TURBODBC_PARAMS)


class EXADialectTurbodbcTest(fixtures.TestBase):
    __skip_if__ = (lambda: testing.db.dialect.driver != 'turbodbc',)

    dialect = None
    default_host_args = {
        'exahost': '192.168.1.2..8:1234',
        'exaschema': 'my_schema'
    }
    default_user_args = {
        'uid': 'scott',
        'pwd': 'tiger',
    }

    @staticmethod
    def _assert_connect_args(result, expected, expected_turbodbc):
        eq_({k: v for k,v in result.items() if k != 'turbodbc_options'}, expected)
        assert (result['turbodbc_options'].read_buffer_size.megabytes
                == expected_turbodbc.get('read_buffer_size', 50))
        assert (result['turbodbc_options'].parameter_sets_to_buffer
                == expected_turbodbc.get('parameter_sets_to_buffer', 1000))
        assert (result['turbodbc_options'].use_async_io
                == expected_turbodbc.get('use_async_io', False))

    def setup(self):
        self.dialect = EXADialect_turbodbc()

    def _get_connection_arguments(self, dsn):
        url = sa_url.make_url(dsn)
        _, args = self.dialect.create_connect_args(url)
        return args

    def assert_parsed(self, dsn, expected_connector, expected_args, expected_turbodbc_args):
        url = sa_url.make_url(dsn)
        connector, args = self.dialect.create_connect_args(url)
        eq_(connector, expected_connector)
        self._assert_connect_args(args, expected_args, expected_turbodbc_args)


    def test_create_connect_args(self):
        expected = self.default_host_args.copy()
        expected.update(self.default_user_args)
        expected.update(DEFAULT_CONNECTION_PARAMS)
        self.assert_parsed("exa+turbodbc://scott:tiger@192.168.1.2..8:1234/my_schema", [None],
                           expected, {})

    def test_create_connect_args_with_driver(self):
        expected = self.default_host_args.copy()
        expected.update(self.default_user_args)
        expected.update(DEFAULT_CONNECTION_PARAMS)
        expected['driver'] = 'EXASolo'
        self.assert_parsed(
            "exa+turbodbc://scott:tiger@192.168.1.2..8:1234/my_schema?driver=EXASolo",
            [None], expected, {})

    def test_create_connect_args_dsn_without_user(self):
        self.assert_parsed("exa+turbodbc://exa_test", ['exa_test'],
                           DEFAULT_CONNECTION_PARAMS, {})

    def test_create_connect_args_dsn(self):
        expected = self.default_user_args.copy()
        expected.update(DEFAULT_CONNECTION_PARAMS)
        self.assert_parsed("exa+pyodbc://scott:tiger@exa_test", ['exa_test'], expected, {})

    def test_create_connect_args_with_turbodbc_args(self):
        expected = self.default_host_args.copy()
        expected.update(self.default_user_args)
        expected.update(DEFAULT_CONNECTION_PARAMS)
        self.assert_parsed("exa+turbodbc://scott:tiger@192.168.1.2..8:1234/my_schema?"
                           "read_buffer_size=4200&parameter_sets_to_buffer=111&use_async_io=True"
                           "&varchar_max_character_limit=99&prefer_unicode=True"
                           "&large_decimals_as_64_bit_types=True"
                           "&limit_varchar_results_to_max=True"
                           "&autocommit=True",
                           [None], expected, {'read_buffer_size': 4200,
                                              'parameter_sets_to_buffer': 111,
                                              'use_async_io': True,
                                              'varchar_max_character_limit': 99,
                                              'prefer_unicode': True,
                                              'large_decimals_as_64_bit_types': True,
                                              'limit_varchar_results_to_max': True,
                                              'autocommit': True})

    def test_create_connect_args_trusted(self):
        expected = self.default_host_args.copy()
        expected.update(DEFAULT_CONNECTION_PARAMS)
        self.assert_parsed("exa+pyodbc://192.168.1.2..8:1234/my_schema", [None], expected, {})

    def test_create_connect_args_with_custom_parameter(self):
        base_url = "exa+pyodbc://scott:tiger@192.168.1.2..8:1234/my_schema"
        assert "custom" not in self._get_connection_arguments(base_url)
        assert self._get_connection_arguments(base_url + "?CUSTOM=something")["custom"] == 'something'

    def test_create_connect_args_with_parameter_set_to_none(self):
        base_url = "exa+pyodbc://scott:tiger@192.168.1.2..8:1234/my_schema"
        assert self._get_connection_arguments(base_url + "?CUSTOM=None")["custom"] is None

    def test_create_connect_args_overrides_default(self):
        base_url = "exa+pyodbc://scott:tiger@192.168.1.2..8:1234/my_schema"

        before = self._get_connection_arguments(base_url)["inttypesinresultsifpossible"]
        after = self._get_connection_arguments(base_url + "?inttypesinresultsifpossible=custom")["inttypesinresultsifpossible"]
        assert before != after



import decimal

from sqlalchemy import types as sqltypes, util

from sqlalchemy_exasol.base import EXADialect


DEFAULT_CONNECTION_PARAMS = {
    # always enable efficient conversion to Python types: 
    # see https://www.exasol.com/support/browse/EXASOL-898
    'inttypesinresultsifpossible': 'y',
}

DEFAULT_TURBODBC_PARAMS = {
    'read_buffer_size': 50
}

TURBODBC_TRANSLATED_PARAMS = {
    'read_buffer_size', 'parameter_sets_to_buffer', 'use_async_io',
    'varchar_max_character_limit', 'prefer_unicode',
    'large_decimals_as_64_bit_types', 'limit_varchar_results_to_max',
    'autocommit'
}


class _ExaDecimal(sqltypes.DECIMAL):
    def bind_processor(self, dialect):
        return super(_ExaDecimal, self).bind_processor(dialect)

    def result_processor(self, dialect, coltype):
        if self.asdecimal:
            fstring = "%%.%df" % self._effective_decimal_return_scale

            def to_decimal(value):
                if value is None:
                    return None
                elif isinstance(value, decimal.Decimal):
                    return value
                elif isinstance(value, float):
                    return decimal.Decimal(fstring % value)
                else:
                    return decimal.Decimal(value)

            return to_decimal
        else:
            return None


class _ExaInteger(sqltypes.INTEGER):
    def bind_processor(self, dialect):
        return super(_ExaInteger, self).bind_processor(dialect)

    def result_processor(self, dialect, coltype):
        def to_integer(value):
            # cast if turbodbc returns a VARCHAR
            if coltype == 30:
                return int(value)
            else:
                return value

        return to_integer


class EXADialect_turbodbc(EXADialect):
    driver = 'turbodbc'
    driver_version = None
    server_version_info = None
    supports_native_decimal = False
    supports_sane_multi_rowcount = False

    colspecs = {sqltypes.Numeric: _ExaDecimal, sqltypes.Integer: _ExaInteger}

    @classmethod
    def dbapi(cls):
        return __import__('turbodbc')

    def create_connect_args(self, url):
        options = self._get_options_with_defaults(url)
        self._translate_none(options)
        self._interpret_destination(options)

        return [[options.pop("dsn", None)], options]

    def _get_server_version_info(self, connection):
        if self.server_version_info is None:
            query = "select PARAM_VALUE from SYS.EXA_METADATA where PARAM_NAME = 'databaseProductVersion'"
            result = connection.execute(query).fetchone()[0].split('.')
            major, minor, patch = 0, 0, 0
            major = int(result[0])
            minor = int(result[1])
            try:
                # last version position can something like: '12-S' or 'RC2'
                # might fail with ValueError
                patch = int(result[2].split('-')[0])
            except ValueError:
                # ignore if there is some funky string in the patch field
                pass
            self.server_version_info = (major, minor, patch)

        # return cached info
        return self.server_version_info

    @staticmethod
    def _get_options_with_defaults(url):
        user_options = url.translate_connect_args(username='uid',
                                                  password='pwd',
                                                  database='exaschema',
                                                  host='destination')
        user_options.update(url.query)

        options = {key.lower(): value for (key, value) in DEFAULT_CONNECTION_PARAMS.items()}
        options.update({key.lower(): value for (key, value) in DEFAULT_TURBODBC_PARAMS.items()})
        for key in user_options.keys():
            options[key.lower()] = user_options[key]

        real_turbodbc = __import__('turbodbc')
        turbodbc_options = {}
        for param in TURBODBC_TRANSLATED_PARAMS:
            if param in options:
                raw = options.pop(param)
                if param in {'use_async_io', 'prefer_unicode',
                             'large_decimals_as_64_bit_types',
                             'limit_varchar_results_to_max',
                             'autocommit'}:
                    value = util.asbool(raw)
                elif param == 'read_buffer_size':
                    value = real_turbodbc.Megabytes(util.asint(raw))
                else:
                    value = util.asint(raw)
                turbodbc_options[param] = value

        options['turbodbc_options'] = real_turbodbc.make_options(**turbodbc_options)

        return options

    @staticmethod
    def _interpret_destination(options):
        if ('port' in options) or ('database' in options):
            options['exahost'] = "{}:{}".format(options.pop('destination'),
                                                options.pop('port'))
        else:
            options['dsn'] = options.pop('destination')

    @staticmethod
    def _translate_none(options):
        for key in options:
            if options[key] == 'None':
                options[key] = None


dialect = EXADialect_turbodbc



# -*- coding: utf-8 -*-
"""Support for the EXASOL database.

Auto Increment Behavior
-----------------------

``IDENTITY`` columns are supported by using SQLAlchemy
``schema.Sequence()`` objects. Example:

    from sqlalchemy import Table, Integer, String, Sequence, Column

    Table('test', metadata,
           Column('id', Integer,
                  Sequence('blah',1000), primary_key=True),
           Column('name', String(20))
         ).create(some_engine)

will yield::

   CREATE TABLE test (
     id INTEGER IDENTITY 1000,
     name VARCHAR(20) NULL,
     PRIMARY KEY(id)
     )

Note that the ``start`` value for sequences is optional and will default to 1.
The start value of a sequence cannot be retrieved when reflecting a Table
object.
The autoincrement flag for Column Objects is not supported by exadialect.

Identifier Casing
-----------------

EXASol mimics the behavior of Oracle. Thus, for this dialect implementation
the Oracle dialect was taken as a reference.
In EXASol, the data dictionary represents all case insensitive identifier names
using UPPERCASE text.SQLAlchemy on the other hand considers an all-lower case
identifiers to be case insensitive. The Oracle dialect converts identifier to
and from those two formats during schema level communication, such as reflection
of tables and indexes.

It is recommended to work with all lowercase identifiers on the SQLAlchemy side.
These are treated as case insensitve identifiers by SQLAlchemy. The EXASol
dialect takes care of converting them to the internal case insensitive
representation (all uppercase).

"""

import six

if six.PY3:
    from six import u as unicode
from decimal import Decimal
from sqlalchemy import sql, schema, types as sqltypes, util, event
from sqlalchemy.schema import AddConstraint, ForeignKeyConstraint
from sqlalchemy.engine import default, reflection, Engine, Connection
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import compiler
from sqlalchemy.sql.elements import quoted_name
from datetime import date, datetime
from .constraints import DistributeByConstraint
import re
import logging

logger=logging.getLogger("sqlalchemy_exasol")

AUTOCOMMIT_REGEXP = re.compile(
    r'\s*(?:UPDATE|INSERT|CREATE|DELETE|DROP|ALTER|TRUNCATE|MERGE)',
    re.I | re.UNICODE)

RESERVED_WORDS = {
    'absolute', 'action', 'add', 'after', 'all', 'allocate', 'alter', 'and', 'any',
    'append', 'are', 'array', 'as', 'asc', 'asensitive', 'assertion', 'at', 'attribute',
    'authid', 'authorization', 'before', 'begin', 'between', 'bigint', 'binary', 'bit',
    'blob', 'blocked', 'bool', 'boolean', 'both', 'by', 'byte', 'call', 'called',
    'cardinality', 'cascade', 'cascaded', 'case', 'casespecific', 'cast', 'catalog',
    'chain', 'char', 'character', 'character_set_catalog', 'character_set_name',
    'character_set_schema', 'characteristics', 'check', 'checked', 'clob', 'close',
    'coalesce', 'collate', 'collation', 'collation_catalog', 'collation_name',
    'collation_schema', 'column', 'commit', 'condition', 'connect_by_iscycle',
    'connect_by_isleaf', 'connect_by_root', 'connection', 'constant', 'constraint',
    'constraint_state_default', 'constraints', 'constructor', 'contains', 'continue',
    'control', 'convert', 'corresponding', 'create', 'cs', 'csv', 'cube', 'current',
    'current_date', 'current_path', 'current_role', 'current_schema', 'current_session',
    'current_statement', 'current_time', 'current_timestamp', 'current_user', 'cursor',
    'cycle', 'data', 'datalink', 'date', 'datetime_interval_code',
    'datetime_interval_precision', 'day', 'dbtimezone', 'deallocate', 'dec', 'decimal',
    'declare', 'default', 'default_like_escape_character', 'deferrable', 'deferred',
    'defined', 'definer', 'delete', 'deref', 'derived', 'desc', 'describe',
    'descriptor', 'deterministic', 'disable', 'disabled', 'disconnect', 'dispatch',
    'distinct', 'dlurlcomplete', 'dlurlpath', 'dlurlpathonly', 'dlurlscheme',
    'dlurlserver', 'dlvalue', 'do', 'domain', 'double', 'drop', 'dynamic',
    'dynamic_function', 'dynamic_function_code', 'each', 'else', 'elseif', 'elsif',
    'emits', 'enable', 'enabled', 'end', 'end-exec', 'enforce', 'equals', 'errors',
    'escape', 'except', 'exception', 'exec', 'execute', 'exists', 'exit', 'export',
    'external', 'extract', 'false', 'fbv', 'fetch', 'file', 'final', 'first', 'float',
    'following', 'for', 'forall', 'force', 'format', 'found', 'free', 'from', 'fs',
    'full', 'function', 'general', 'generated', 'geometry', 'get', 'global', 'go',
    'goto', 'grant', 'granted', 'group', 'groups', 'group_concat', 'grouping', 'having',
    'high', 'hold', 'hour', 'identity', 'if', 'ifnull', 'immediate', 'implementation',
    'import', 'in', 'index', 'indicator', 'inner', 'inout', 'input', 'insensitive',
    'insert', 'instance', 'instantiable', 'int', 'integer', 'integrity', 'intersect',
    'interval', 'into', 'inverse', 'invoker', 'is', 'iterate', 'join', 'key_member',
    'key_type', 'large', 'last', 'lateral', 'ldap', 'leading', 'leave', 'left', 'level',
    'like', 'limit', 'listagg', 'local', 'localtime', 'localtimestamp', 'locator',
    'log', 'longvarchar', 'loop', 'low', 'map', 'match', 'matched', 'merge', 'method',
    'minus', 'minute', 'mod', 'modifies', 'modify', 'module', 'month', 'names',
    'national', 'natural', 'nchar', 'nclob', 'new', 'next', 'nls_date_format',
    'nls_date_language', 'nls_first_day_of_week', 'nls_numeric_characters',
    'nls_timestamp_format', 'no', 'nocycle', 'nologging', 'none', 'not', 'null',
    'nullif', 'number', 'numeric', 'nvarchar', 'nvarchar2', 'object', 'of', 'off',
    'old', 'on', 'only', 'open', 'option', 'options', 'or', 'order', 'ordering',
    'ordinality', 'others', 'out', 'outer', 'output', 'over', 'overlaps', 'overlay',
    'overriding', 'pad', 'parallel_enable', 'parameter', 'parameter_specific_catalog',
    'parameter_specific_name', 'parameter_specific_schema', 'partial', 'path',
    'permission', 'placing', 'plus', 'position', 'preceding', 'preferring', 'prepare',
    'preserve', 'prior', 'privileges', 'procedure', 'profile', 'random', 'range',
    'read', 'reads', 'real', 'recovery', 'recursive', 'ref', 'references',
    'referencing', 'refresh', 'regexp_like', 'relative', 'release', 'rename', 'repeat',
    'replace', 'restore', 'restrict', 'result', 'return', 'returned_length',
    'returned_octet_length', 'returns', 'revoke', 'right', 'rollback', 'rollup',
    'routine', 'row', 'rows', 'rowtype', 'savepoint', 'schema', 'scope', 'script',
    'scroll', 'search', 'second', 'section', 'security', 'select', 'selective', 'self',
    'sensitive', 'separator', 'sequence', 'session', 'session_user', 'sessiontimezone',
    'set', 'sets', 'shortint', 'similar', 'smallint', 'some', 'source', 'space',
    'specific', 'specifictype', 'sql', 'sql_bigint', 'sql_bit', 'sql_char', 'sql_date',
    'sql_decimal', 'sql_double', 'sql_float', 'sql_integer', 'sql_longvarchar',
    'sql_numeric', 'sql_preprocessor_script', 'sql_real', 'sql_smallint',
    'sql_timestamp', 'sql_tinyint', 'sql_type_date', 'sql_type_timestamp',
    'sql_varchar', 'sqlexception', 'sqlstate', 'sqlwarning', 'start', 'state',
    'statement', 'static', 'structure', 'style', 'substring', 'subtype', 'sysdate',
    'system', 'system_user', 'systimestamp', 'table', 'temporary', 'text', 'then',
    'time', 'timestamp', 'timezone_hour', 'timezone_minute', 'tinyint', 'to',
    'trailing', 'transaction', 'transform', 'transforms', 'translation', 'treat',
    'trigger', 'trim', 'true', 'truncate', 'under', 'union', 'unique', 'unknown',
    'unlink', 'unnest', 'until', 'update', 'usage', 'user', 'using', 'value', 'values',
    'varchar', 'varchar2', 'varray', 'verify', 'view', 'when', 'whenever', 'where',
    'while', 'window', 'with', 'within', 'without', 'work', 'year', 'yes', 'zone'
}

colspecs = {
}

ischema_names = {
    'BOOLEAN': sqltypes.BOOLEAN,
    'CHAR': sqltypes.CHAR,
    'CLOB': sqltypes.TEXT,
    'DATE': sqltypes.DATE,
    'DECIMAL': sqltypes.DECIMAL,
    'DOUBLE': sqltypes.FLOAT,  # EXASOL mapps DOUBLE, DOUBLE PRECISION, FLOAT to DOUBLE PRECISION
    # internally but returns 'DOUBLE' as type when asking the DB catalog
    # INTERVAL DAY [(p)] TO SECOND [(fp)] TODO: missing support for EXA Datatype, check Oracle Engine
    # INTERVAL YEAR[(p)] TO MONTH         TODO: missing support for EXA Datatype, check Oracle Engine
    'TIMESTAMP': sqltypes.TIMESTAMP,
    'VARCHAR': sqltypes.VARCHAR,
}


class EXACompiler(compiler.SQLCompiler):
    extract_map = util.update_copy(
        compiler.SQLCompiler.extract_map,
        {
            'month': '%m',
            'day': '%d',
            'year': '%Y',
            'second': '%S',
            'hour': '%H',
            'doy': '%j',
            'minute': '%M',
            'epoch': '%s',
            'dow': '%w',
            'week': '%W'
        })

    def visit_now_func(self, fn, **kw):
        return "CURRENT_TIMESTAMP"

    def visit_char_length_func(self, fn, **kw):
        return "length%s" % self.function_argspec(fn)

    def limit_clause(self, select, **kw):
        text = ""
        if select._limit is not None:
            text += "\n LIMIT %d" % int(select._limit)
        if select._offset is not None:
            text += "\n OFFSET %d" % int(select._offset)

        return text

    def for_update_clause(self, select):
        # Exasol has no "FOR UPDATE"
        util.warn("EXASolution does not support SELECT ... FOR UPDATE")
        return ''

    def default_from(self):
        """Called when a ``SELECT`` statement has no froms,
        and no ``FROM`` clause is to be appended.
        """
        return " FROM DUAL"

    def visit_empty_set_expr(self, type_):
        return "SELECT 1 FROM DUAL WHERE 1!=1"


class EXADDLCompiler(compiler.DDLCompiler):

    def get_column_specification(self, column, **kwargs):
        colspec = self.preparer.format_column(column)
        colspec += " " + self.dialect.type_compiler.process(column.type)
        if column is column.table._autoincrement_column and \
                True and \
                (
                        column.default is None or \
                        isinstance(column.default, schema.Sequence)
                ):
            colspec += " IDENTITY"
            if isinstance(column.default, schema.Sequence) and \
                    column.default.start > 0:
                colspec += " " + str(column.default.start)
        else:
            default = self.get_column_default_string(column)
            if default is not None:
                colspec += " DEFAULT " + default

        if not column.nullable:
            colspec += " NOT NULL"
        return colspec

    def create_table_constraints(self, table, _include_foreign_key_constraints=None):
        # EXASOL does not support FK constraints that reference
        # the table being created. Thus, these need to be created
        # via ALTER TABLE after table creation
        # TODO: FKs that reference other tables could be inlined
        # the create rule could be more specific but for now, ALTER
        # TABLE for all FKs work.
        for c in [c for c in table._sorted_constraints if isinstance(c, ForeignKeyConstraint)]:
            c._create_rule = lambda: False
            event.listen(table, "after_create", AddConstraint(c))
        return super(EXADDLCompiler, self).create_table_constraints(table)

    def visit_add_constraint(self, create):
        if isinstance(create.element, DistributeByConstraint):
            return "ALTER TABLE %s %s" % (
                self.preparer.format_table(create.element.table),
                self.process(create.element)
            )
        else:
            return super(EXADDLCompiler, self).visit_add_constraint(create)

    def visit_drop_constraint(self, drop):
        if isinstance(drop.element, DistributeByConstraint):
            return "ALTER TABLE %s DROP DISTRIBUTION KEYS" % (
                self.preparer.format_table(drop.element.table))
        else:
            return super(EXADDLCompiler, self).visit_drop_constraint(drop)

    def visit_distribute_by_constraint(self, constraint):
        return "DISTRIBUTE BY " + ",".join(c.name for c in constraint.columns)

    def define_constraint_remote_table(self, constraint, table, preparer):
        """Format the remote table clause of a CREATE CONSTRAINT clause."""
        return preparer.format_table(table, use_schema=True)

    def visit_create_index(self, create):
        """EXASol manages indexes internally"""
        raise NotImplementedError()

    def visit_drop_index(self, drop):
        """EXASol manages indexes internally"""
        raise NotImplementedError()


class EXATypeCompiler(compiler.GenericTypeCompiler):
    """ force mapping of BIGINT to DECIMAL(19)
    The mapping back is done by the driver using flag
    INTTYPESINRESULTSIFPOSSIBLE=Y. This is enforced by default by this
    dialect. However, BIGINT is mapped to DECIMAL(36) and the driver only
    converts types decimal scale==0 and 9<precision<=19 back to BIGINT
    https://www.exasol.com/support/browse/EXA-23267
    """

    def visit_big_integer(self, type_, **kw):
        return "DECIMAL(19)"

    def visit_large_binary(self, type_):
        return self.visit_BLOB(type_)

    def visit_datetime(self, type_):
        return self.visit_TIMESTAMP(type_)


class EXAIdentifierPreparer(compiler.IdentifierPreparer):
    reserved_words = RESERVED_WORDS
    illegal_initial_characters = compiler.ILLEGAL_INITIAL_CHARACTERS.union('_')


class EXAExecutionContext(default.DefaultExecutionContext):

    def fire_sequence(self, default, type_):
        raise NotImplemented

    def get_insert_default(self, column):
        if column.default.is_sequence:
            return 'DEFAULT'
        else:
            return super(EXAExecutionContext, self).get_insert_default(column)

    def get_lastrowid(self):
        columns = self.compiled.sql_compiler.statement.table.columns
        autoinc_pk_columns = \
            [c.name for c in columns if c.autoincrement and c.primary_key]
        if len(autoinc_pk_columns) == 0:
            return None
        elif len(autoinc_pk_columns) > 1:
            util.warn("Table with more than one autoincrement, primary key" \
                      " Column!")
            raise Exception
        else:
            id_col = self.dialect.denormalize_name(autoinc_pk_columns[0])

            table = self.compiled.sql_compiler.statement.table.name
            table = self.dialect.denormalize_name(table)

            sql_stmnt = "SELECT column_identity from SYS.EXA_ALL_COLUMNS " \
                        "WHERE column_object_type = 'TABLE' and column_table " \
                        "= ? AND column_name = ?"

            schema = self.compiled.sql_compiler.statement.table.schema
            if schema is not None:
                schema = self.dialect.denormalize_name(schema)
                sql_stmnt += " AND column_schema = ?"

            cursor = self.create_cursor()
            if schema:
                cursor.execute(sql_stmnt, (table, id_col, schema))
            else:
                cursor.execute(sql_stmnt, (table, id_col))
            result = cursor.fetchone()
            cursor.close()
            return int(result[0]) - 1

    def pre_exec(self):
        """
        This routine inserts the parameters into the compiled query prior to executing it.
        The reason for this workaround is the poor performance for prepared statements.
        Note: Parameter replacement is done for server versions < 4.1.8 or
              in case a delete query is executed.
        """
        if self.isdelete or self.root_connection.dialect.server_version_info < (4, 1, 8):
            db_query = self.unicode_statement
            for i in range(1, len(self.parameters)):
                db_query += ", (" + ", ".join(['?'] * len(self.parameters[i])) + ")"
            for db_para in self.parameters:
                for value in db_para:
                    ident = '?'
                    if value is None:
                        db_query = db_query.replace(ident, 'NULL', 1)
                    elif isinstance(value, six.integer_types):
                        db_query = db_query.replace(ident, str(value), 1)
                    elif isinstance(value, (float, Decimal)):
                        db_query = db_query.replace(ident, str(float(value)), 1)
                    elif isinstance(value, bool):
                        db_query = db_query.replace(ident, '1' if value else '0', 1)
                    elif isinstance(value, datetime):
                        db_query = db_query.replace(ident,
                                                    "to_timestamp('%s', 'YYYY-MM-DD HH24:MI:SS.FF6')" % value.strftime(
                                                        '%Y-%m-%d %H:%M:%S.%f'), 1)
                    elif isinstance(value, date):
                        db_query = db_query.replace(ident, "to_date('%s', 'YYYY-MM-DD')" % value.strftime('%Y-%m-%d'),
                                                    1)
                    elif isinstance(value, six.binary_type):
                        db_query = db_query.replace(ident, "'%s'" % value.decode('UTF-8'), 1)
                    elif isinstance(value, six.text_type):
                        db_query = db_query.replace(ident, "'%s'" % value, 1)
                    else:
                        raise TypeError('Data type not supported: %s' % type(value))
            self.statement = db_query
            self.parameters = [[]]

    def should_autocommit_text(self, statement):
        return AUTOCOMMIT_REGEXP.match(statement)


class EXADialect(default.DefaultDialect):
    name = 'exasol'
    supports_native_boolean = True
    supports_native_decimal = True
    supports_alter = True
    supports_unicode_statements = True
    supports_unicode_binds = True
    supports_default_values = True
    supports_empty_insert = False
    supports_sequences = False
    supports_is_distinct_from = False
    # sequences_optional = True
    # controls in SQLAlchemy base which columns are part of an insert statement
    # postfetch_lastrowid = True
    supports_cast = True
    requires_name_normalize = True
    cte_follows_insert = True

    default_paramstyle = 'qmark'
    execution_ctx_cls = EXAExecutionContext
    statement_compiler = EXACompiler
    ddl_compiler = EXADDLCompiler
    type_compiler = EXATypeCompiler
    preparer = EXAIdentifierPreparer
    ischema_names = ischema_names
    colspecs = colspecs
    isolation_level = None

    def __init__(self, isolation_level=None, native_datetime=False, **kwargs):
        default.DefaultDialect.__init__(self, **kwargs)
        self.isolation_level = isolation_level

    _isolation_lookup = {
        'SERIALIZABLE': 0
    }

    def _get_default_schema_name(self, connection):
        """
        Default schema is derived from current connections url
        or 'SYS'. Tables in 'SYS' are not reflectable!
        """
        schema = None
        schema = self._get_schema_from_url(connection, schema)
        if schema is None:
            schema = self.normalize_name(u"SYS")
        return schema

    def _get_schema_from_url(self, connection, schema):
        if connection.engine.url is not None and connection.engine.url != "":
            schema = self.normalize_name(
                connection.engine.url.translate_connect_args().get('database'))
        return schema

    def normalize_name(self, name):
        """
        Converting EXASol case insensitive identifiers (upper case)
        to  SQLAlchemy case insensitive identifiers (lower case)
        """
        if name is None:
            return None

        if six.PY2:
            if isinstance(name, str):
                name = name.decode(self.encoding)

        if name.upper() == name and \
                not self.identifier_preparer._requires_quotes(name.lower()):
            return name.lower()
        elif name.lower() == name:
            return quoted_name(name, quote=True)
        else:
            return name

    def denormalize_name(self, name):
        """
        Converting SQLAlchemy case insensitive identifiers (lower case)
        to  EXASol case insensitive identifiers (upper case)
        """
        if name is None or len(name) == 0:
            return None
        elif name.lower() == name and \
                not self.identifier_preparer._requires_quotes(name.lower()):
            name = name.upper()

        if six.PY2:
            if not self.supports_unicode_binds:
                name = name.encode(self.encoding)
            else:
                name = unicode(name)

        return name

    def get_isolation_level(self, connection):
        return "SERIALIZABLE"

    def on_connect(self):
        # TODO: set isolation level
        pass

    def getODBCConnection(self, connection):
        if isinstance(connection, Engine):
            odbc_connection = connection.raw_connection().connection
        elif isinstance(connection, Session):
            odbc_connection = connection.connection()
            return self.getODBCConnection(odbc_connection)
        elif isinstance(connection, Connection):
            odbc_connection = connection.connection.connection
        else:
            return None
        if "pyodbc.Connection" in str(type(odbc_connection)):
            return odbc_connection
        else:
            return None

    def use_sql_fallback(self, **kw):
        result = "use_sql_fallback" in kw and kw.get("use_sql_fallback") == True
        if result:
            logger.warning("Using sql fallback instead of odbc functions")
        return result

    # never called during reflection
    @reflection.cache
    def get_schema_names(self, connection, **kw):
        if self.use_sql_fallback(**kw):
            prefix = "/*snapshot execution*/ "
        else:
            prefix = ""
        sql_stmnt = "%sselect SCHEMA_NAME from SYS.EXA_SCHEMAS" % prefix
        rs = connection.execute(sql.text(sql_stmnt))
        return [self.normalize_name(row[0]) for row in rs]

    def _get_schema_for_input_or_current(self, connection, schema):
        schema = self._get_schema_for_input(connection, schema)
        if schema is None:
            schema = self._get_current_schema(connection)
        return self.denormalize_name(schema)

    def _get_schema_for_input(self, connection, schema):
        schema = self.denormalize_name(schema or self._get_schema_from_url(connection, schema))
        return schema

    def _get_current_schema(self, connection):
        current_schema_stmnt = "SELECT CURRENT_SCHEMA"
        current_schema = connection.execute(current_schema_stmnt).fetchone()[0]
        return current_schema

    @reflection.cache
    def _get_tables_for_schema_odbc(self, connection, odbc_connection, schema, table_type=None, table_name=None, **kw):
        schema = self._get_schema_for_input_or_current(connection, schema)
        table_name = self.denormalize_name(table_name)
        with odbc_connection.cursor().tables(schema=schema, tableType=table_type, table=table_name) as table_cursor:
            rows = [row for row in table_cursor]
            return rows

    @reflection.cache
    def get_table_names(self, connection, schema, **kw):
        odbc_connection = self.getODBCConnection(connection)
        if odbc_connection is not None and not self.use_sql_fallback(**kw):
            return self.get_table_names_odbc(connection, odbc_connection, schema, **kw)
        else:
            return self.get_table_names_sql(connection, schema, **kw)

    @reflection.cache
    def get_table_names_odbc(self, connection, odbc_connection, schema, **kw):
        tables = self._get_tables_for_schema_odbc(connection, odbc_connection, schema, table_type="TABLE", **kw)
        normalized_tables = [self.normalize_name(row.table_name)
                             for row in tables]
        return normalized_tables

    @reflection.cache
    def get_table_names_sql(self, connection, schema, **kw):
        schema = self._get_schema_for_input(connection, schema)
        sql_stmnt = "SELECT table_name FROM  SYS.EXA_ALL_TABLES WHERE table_schema = "
        if schema is None:
            sql_stmnt += "CURRENT_SCHEMA ORDER BY table_name"
            rs = connection.execute(sql_stmnt)
        else:
            sql_stmnt += ":schema ORDER BY table_name"
            rs = connection.execute(sql.text(sql_stmnt),
                                    schema=self.denormalize_name(schema))
        tables = [self.normalize_name(row[0]) for row in rs]
        return tables

    def has_table(self, connection, table_name, schema=None, **kw):
        odbc_connection = self.getODBCConnection(connection)
        if odbc_connection is not None and not self.use_sql_fallback(**kw):
            result=self.has_table_odbc(connection, odbc_connection, schema=schema, table_name=table_name, **kw)
        else:
            result=self.has_table_sql(connection, schema=schema, table_name=table_name, **kw) 
        return result

    def has_table_odbc(self, connection, odbc_connection, table_name, schema=None, **kw):
        tables = self.get_table_names_odbc(connection=connection,
                                           odbc_connection=odbc_connection,
                                           schema=schema, table_name=table_name, **kw)
        result = self.normalize_name(table_name) in tables
        return result

    def has_table_sql(self, connection, table_name, schema=None, **kw):
        schema = self._get_schema_for_input(connection, schema)
        sql_stmnt = "SELECT table_name from SYS.EXA_ALL_TABLES " \
                    "WHERE table_name = :table_name "
        if schema is not None:
            sql_stmnt += "AND table_schema = :schema"
        rp = connection.execute(
            sql.text(sql_stmnt),
            table_name=self.denormalize_name(table_name),
            schema=self.denormalize_name(schema))
        row = rp.fetchone()

        return (row is not None)

    @reflection.cache
    def get_view_names(self, connection, schema=None, **kw):
        odbc_connection = self.getODBCConnection(connection)
        if odbc_connection is not None and not self.use_sql_fallback(**kw):
            return self.get_view_names_odbc(connection, odbc_connection, schema, **kw)
        else:
            return self.get_view_names_sql(connection, schema, **kw)

    @reflection.cache
    def get_view_names_odbc(self, connection, odbc_connection, schema=None, **kw):
        tables = self._get_tables_for_schema_odbc(connection, odbc_connection, schema, table_type="VIEW", **kw)
        return [self.normalize_name(row.table_name)
                for row in tables]

    def get_view_names_sql(self, connection, schema=None, **kw):
        schema = self._get_schema_for_input(connection, schema)
        sql_stmnt = "SELECT view_name FROM  SYS.EXA_ALL_VIEWS WHERE view_schema = "
        if schema is None:
            sql_stmnt += "CURRENT_SCHEMA ORDER BY view_name"
            rs = connection.execute(sql.text(sql_stmnt))
        else:
            sql_stmnt += ":schema ORDER BY view_name"
            rs = connection.execute(sql.text(sql_stmnt),
                                    schema=self.denormalize_name(schema))
        return [self.normalize_name(row[0]) for row in rs]

    @reflection.cache
    def get_view_definition(self, connection, view_name, schema=None, **kw):
        odbc_connection = self.getODBCConnection(connection)
        if odbc_connection is not None and not self.use_sql_fallback(**kw):
            return self.get_view_definition_odbc(connection, odbc_connection, view_name, schema, **kw)
        else:
            return self.get_view_definition_sql(connection, view_name, schema, **kw)

    def quote_string_value(self, string_value):
        return "'%s'" % (string_value.replace("'", "''"))

    @reflection.cache
    def get_view_definition_odbc(self, connection, odbc_connection, view_name, schema=None, **kw):
        if view_name is None:
            return None
        tables = self._get_tables_for_schema_odbc(connection, odbc_connection, schema, table_type="VIEW",
                                                  table_name=view_name, **kw)
        if len(tables) == 1:
            quoted_view_name_string = self.quote_string_value(tables[0][2])
            quoted_view_schema_string = self.quote_string_value(tables[0][1])
            sql_stmnt = \
                "/*snapshot execution*/ SELECT view_text FROM sys.exa_all_views WHERE view_name = {view_name} AND view_schema = {view_schema}".format(
                    view_name=quoted_view_name_string, view_schema=quoted_view_schema_string)
            rp = connection.execute(sql.text(sql_stmnt)).scalar()
            if rp:
                if six.PY3:
                    return rp
                else:
                    return rp.decode(self.encoding)
            else:
                return None
        else:
            return None

    def get_view_definition_sql(self, connection, view_name, schema=None, **kw):
        schema = self._get_schema_for_input(connection, schema)
        sql_stmnt = "SELECT view_text FROM sys.exa_all_views WHERE view_name = :view_name AND view_schema = "
        if schema is None:
            sql_stmnt += "CURRENT_SCHEMA"
        else:
            sql_stmnt += ":schema"
        rp = connection.execute(sql.text(sql_stmnt),
                                view_name=self.denormalize_name(view_name),
                                schema=self.denormalize_name(schema)).scalar()
        if rp:
            if six.PY3:
                return rp
            else:
                return rp.decode(self.encoding)
        else:
            return None

    def get_column_sql_query_str(self):
        return "SELECT " \
               "column_name, " \
               "column_type, " \
               "column_maxsize, " \
               "column_num_prec, " \
               "column_num_scale, " \
               "column_is_nullable, " \
               "column_default, " \
               "column_identity, " \
               "column_is_distribution_key, " \
               "column_table " \
               "FROM sys.exa_all_columns " \
               "WHERE " \
               "column_object_type IN ('TABLE', 'VIEW') AND " \
               "column_schema = {schema} AND " \
               "column_table = {table} " \
               "ORDER BY column_ordinal_position"

    @reflection.cache
    def _get_columns_odbc(self, connection, odbc_connection, table_name, schema, **kw):
        tables = self._get_tables_for_schema_odbc(connection, odbc_connection,
                                                  schema=schema, table_name=table_name, **kw)
        if len(tables) == 1:
            # get_columns_sql originally returned all columns of all tables if table_name is None,
            # we follow this behavior here for compatibility. However, the documentation for Dialects
            # does not mentions this behavior:
            # https://docs.sqlalchemy.org/en/13/core/internals.html#sqlalchemy.engine.interfaces.Dialect
            quoted_schema_string = self.quote_string_value(tables[0].table_schem)
            quoted_table_string = self.quote_string_value(tables[0].table_name)
            sql_stmnt = \
                "/*snapshot execution*/ " + \
                self.get_column_sql_query_str() \
                    .format(schema=quoted_schema_string, table=quoted_table_string)
            rp = connection.execute(sql.text(sql_stmnt))
            return list(rp)
        else:
            return []

    @reflection.cache
    def _get_columns_sql(self, connection, table_name, schema=None, **kw):
        schema = self._get_schema_for_input(connection, schema)
        if schema is None:
            schema_str = "CURRENT_SCHEMA"
        else:
            schema_str = ":schema"
        table_name_str = ":table"
        sql_stmnt = \
            self.get_column_sql_query_str() \
                .format(schema=schema_str, table=table_name_str)
        stmnt = sql.text(sql_stmnt)
        rp = connection.execute(stmnt, 
                schema=self.denormalize_name(schema),
                table=self.denormalize_name(table_name))

        return list(rp)

    @reflection.cache
    def _get_columns(self, connection, table_name, schema=None, **kw):
        odbc_connection = self.getODBCConnection(connection)
        if odbc_connection is not None and not self.use_sql_fallback(**kw):
            columns = self._get_columns_odbc(connection, odbc_connection, table_name, schema, **kw)
        else:
            columns = self._get_columns_sql(connection, table_name, schema, **kw)
        return columns

    @reflection.cache
    def get_columns(self, connection, table_name, schema=None, **kw):
        if table_name is None:
            return []

        columns = []
        rows = self._get_columns(connection,
                                 table_name=table_name,
                                 schema=schema, **kw)
        table_name = self.denormalize_name(table_name)
        for row in rows:
            (colname, coltype, length, precision, scale, nullable, default, identity, is_distribution_key) = \
                (row[0], row[1], row[2], row[3], row[4], row[5], row[6], row[7], row[8])

            # FIXME: Missing type support: INTERVAL DAY [(p)] TO SECOND [(fp)], INTERVAL YEAR[(p)] TO MONTH

            # remove ASCII, UTF8 and spaces from char-like types
            coltype = re.sub(r'ASCII|UTF8| ', '', coltype)
            # remove precision and scale addition from numeric types
            coltype = re.sub(r'\(\d+(\,\d+)?\)', '', coltype)
            try:
                if coltype == 'VARCHAR':
                    coltype = sqltypes.VARCHAR(length)
                elif coltype == 'CHAR':
                    coltype = sqltypes.CHAR(length)
                elif coltype == 'DECIMAL':
                    # this Dialect forces INTTYPESINRESULTSIFPOSSIBLE=y on ODBC level
                    # thus, we need to convert DECIMAL(<=18,0) back to INTEGER type
                    # and DECIMAL(36,0) back to BIGINT type
                    if scale == 0 and precision <= 18:
                        coltype = sqltypes.INTEGER()
                    elif scale == 0 and precision == 36:
                        coltype = sqltypes.BIGINT()
                    else:
                        coltype = sqltypes.DECIMAL(precision, scale)
                else:
                    coltype = self.ischema_names[coltype]
            except KeyError:
                util.warn("Did not recognize type '%s' of column '%s'" %
                          (coltype, colname))
                coltype = sqltypes.NULLTYPE

            cdict = {
                'name': self.normalize_name(colname),
                'type': coltype,
                'nullable': nullable,
                'default': default,
                'is_distribution_key': is_distribution_key
            }
            if identity:
                identity = int(identity)
            # if we have a positive identity value add a sequence
            if identity is not None and identity >= 0:
                cdict['sequence'] = {'name': ''}
                # TODO: we have to possibility to encode the current identity value count
                # into the column metadata. But the consequence is that it would also be used
                # as start value in CREATE statements. For now the current value is ignored.
                # Add it by changing the dict to: {'name':'', 'start': int(identity)}
            columns.append(cdict)
        return columns

    def _get_constraint_sql_str(self, schema, table_name, contraint_type):
        sql_stmnt = \
            "SELECT " \
            "constraint_name, " \
            "column_name, " \
            "referenced_schema, " \
            "referenced_table, " \
            "referenced_column, " \
            "constraint_table, " \
            "constraint_type " \
            "FROM SYS.EXA_ALL_CONSTRAINT_COLUMNS " \
            "WHERE " \
            "constraint_schema={schema} AND " \
            "constraint_table={table_name} AND " \
            "constraint_type='{contraint_type}' " \
            "ORDER BY ordinal_position" \
                .format(schema=schema, table_name=table_name, contraint_type=contraint_type)
        return sql_stmnt

    @reflection.cache
    def get_pk_constraint_odbc(self, connection, odbc_connection, table_name, schema=None, **kw):
        schema = self._get_schema_for_input_or_current(connection, schema)
        table_name = self.denormalize_name(table_name)
        with odbc_connection.cursor().primaryKeys(table=table_name, schema=schema) as primaryKeys_cursor:
            pkeys = []
            constraint_name = None
            for row in primaryKeys_cursor:
                if row[2] != table_name and table_name is not None:
                    continue
                pkeys.append(self.normalize_name(row[3]))
                constraint_name = self.normalize_name(row[5])
        return {'constrained_columns': pkeys, 'name': constraint_name}

    @reflection.cache
    def get_pk_constraint_sql(self, connection, table_name, schema=None, **kw):
        schema = self._get_schema_for_input(connection, schema)
        table_name = self.denormalize_name(table_name)
        table_name_string = ":table"
        if schema is None:
            schema_string = "CURRENT_SCHEMA "
        else:
            schema_string = ":schema "
        sql_stmnt=self._get_constraint_sql_str(schema_string, table_name_string, "PRIMARY KEY")
        rp = connection.execute(sql.text(sql_stmnt),
                                schema=self.denormalize_name(schema),
                                table=table_name)
        pkeys = []
        constraint_name = None
        for row in list(rp):
            if (row[5] != table_name and table_name is not None) or row[6] != 'PRIMARY KEY':
                continue
            pkeys.append(self.normalize_name(row[1]))
            constraint_name = self.normalize_name(row[0])
        return {'constrained_columns': pkeys, 'name': constraint_name}


    @reflection.cache
    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        if table_name is None:
            return None
        odbc_connection = self.getODBCConnection(connection)
        if odbc_connection is not None and not self.use_sql_fallback(**kw):
            return self.get_pk_constraint_odbc(connection, odbc_connection, table_name=table_name, schema=schema, **kw)
        else:
            return self.get_pk_constraint_sql(connection, table_name=table_name, schema=schema, **kw)

    @reflection.cache
    def _get_foreign_keys_odbc(self, connection, odbc_connection, table_name, schema=None, **kw):
        # Need to use a workaround, because SQLForeignKeys functions doesn't work for an unknown reason
        tables = self._get_tables_for_schema_odbc(connection=connection, odbc_connection=odbc_connection,
                                                  schema=schema, table_name=table_name, table_type="TABLE", **kw)
        if len(tables) > 0:
            quoted_schema_string = self.quote_string_value(tables[0].table_schem)
            quoted_table_string = self.quote_string_value(tables[0].table_name)
            sql_stmnt = \
                "/*snapshot execution*/ " + \
                self._get_constraint_sql_str(quoted_schema_string,quoted_table_string,"FOREIGN KEY")
            rp = connection.execute(sql.text(sql_stmnt))
            return list(rp)
        else:
            return []

    @reflection.cache
    def _get_foreign_keys_sql(self, connection, table_name, schema=None, **kw):
        table_name_string = ":table"
        if schema is None:
            schema_string = "CURRENT_SCHEMA "
        else:
            schema_string = ":schema "
        sql_stmnt = \
            self._get_constraint_sql_str(schema_string, table_name_string, "FOREIGN KEY")
        rp = connection.execute(sql.text(sql_stmnt),
                                schema=self.denormalize_name(schema),
                                table=self.denormalize_name(table_name))
        return list(rp)

    @reflection.cache
    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        if table_name is None:
            return []
        schema_int = self._get_schema_for_input_or_current(connection, schema)

        def fkey_rec():
            return {
                'name': None,
                'constrained_columns': [],
                'referred_schema': None,
                'referred_table': None,
                'referred_columns': []
            }

        fkeys = util.defaultdict(fkey_rec)
        odbc_connection = self.getODBCConnection(connection)
        if odbc_connection is not None and not self.use_sql_fallback(**kw):
            constraints = self._get_foreign_keys_odbc(connection, odbc_connection, table_name=table_name,
                                                      schema=schema_int, **kw)
        else:
            constraints = self._get_foreign_keys_sql(connection, table_name=table_name, schema=schema_int, **kw)
        table_name = self.denormalize_name(table_name)
        for row in constraints:
            (cons_name, local_column, remote_schema, remote_table, remote_column) = \
                (row[0], row[1], row[2], row[3], row[4])
            rec = fkeys[self.normalize_name(cons_name)]
            rec['name'] = self.normalize_name(cons_name)
            local_cols, remote_cols = rec['constrained_columns'], rec['referred_columns']

            if not rec['referred_table']:
                rec['referred_table'] = self.normalize_name(remote_table)
                # we need to take care of calls without schema. the sqla test suite
                # expects referred_schema to be None if None is passed in to this function
                if schema is None and self.normalize_name(schema_int) == self.normalize_name(remote_schema):
                    rec['referred_schema'] = None
                else:
                    rec['referred_schema'] = self.normalize_name(remote_schema)

            local_cols.append(self.normalize_name(local_column))
            remote_cols.append(self.normalize_name(remote_column))

        result = list(fkeys.values())
        return result

    @reflection.cache
    def get_indexes(self, connection, table_name, schema=None, **kw):
        # EXASolution has no explicit indexes
        return []
